! Module deals with particle ricochets
! Contains functions for reading and writing parameter values
! And master ricochet function that can choose any number of sub functions

module ricochet
  use sysdep_io
  use precision
  use maths_const
  use grain_var
  use random
  use vector
  use string_param
  use random_distrib
  use maths_fnct
  private

  public :: read_ricochet_nl    
  public :: write_ricochet_nl    
  public :: init_ricochet    
  public :: end_ricochet    
  public :: grain_ricochet       ! Calculate the results of a ricochet

  integer :: decoded_type = 0 ! Decode the string for the function type
  integer, parameter :: FLAT_T        = 1
  integer, parameter :: IAN_T         = 2
  integer, parameter :: THEORETICAL_T = 3
  integer, parameter :: FIXED_T       = 4
  integer, parameter :: TEIONKEN_T    = 5

  ! flat - like bouncing of a flat plate with 
  !        collision determined by restitution distribution, friction
  !        ejection angle gets offset by angle distribution
  !
  ! fixed - ejection is independent of impact angle
  !       - restitution and angle determined by fixed distribution 
        
  ! List of known function types
  type(param_table_t), parameter :: types(5) = (/&
       param_table_t("flat",       FLAT_T),&  
       param_table_t("ian",        IAN_T),&
       param_table_t("theoretical",THEORETICAL_T),&
       param_table_t("teionken",   TEIONKEN_T),&
       param_table_t("fixed",      FIXED_T)/)

  ! string describing type of ricochet function
  character(len=STRING_LENGTH) :: ricochet_type='fixed' 
  real(r_kind) :: friction    = .3    ! contact friction
  type(fnct_pub_t)    :: rico_prob    = fnct_pub_t("heaviside",(/0.,.94,5.,0./))
  type(distrib_pub_t) :: angle        = distrib_pub_t("normal",38,10)
  type(distrib_pub_t) :: restitution  = distrib_pub_t("normal",0.56,0.2)

  namelist /ricochet_nl/ ricochet_type, friction, rico_prob, angle, restitution

  type(fnct_t)    :: prob_fnct ! Fucntion for probability of ricochet
  type(distrib_t) :: ang_dist  ! Angle distribution
  type(distrib_t) :: e_dist    ! Restitution distribution
 
contains
  subroutine read_ricochet_nl(fp)
    integer, intent(in) :: fp
    read(fp,nml=ricochet_nl)
  end subroutine read_ricochet_nl
  subroutine write_ricochet_nl(fp)
    integer, intent(in) :: fp  
    write(fp,'(A)')      ' &ricochet_nl'
    write(fp,'(A)')      '    ricochet_type = "' // trim(ricochet_type) // '"'
    write(fp,'(A,G11.4)')'    friction      = ', friction
    write(fp,'(A,G11.4)',advance='no')'    rico_prob     = '
    call write_fnct(fp,rico_prob)
    write(fp,'(A,G11.4)',advance='no')'    angle         = '
    call write_distrib(fp,angle)
    write(fp,'(A,G11.4)',advance='no')'    restitution   = '
    call write_distrib(fp,restitution)
    write(fp,'(A)')      ' /'
    write(fp,'(A)')' '
  end subroutine write_ricochet_nl
  subroutine init_ricochet()
    type(distrib_pub_t) a
    decoded_type = decode_string_param(ricochet_type,types)

    ! Convert angles to a scale of [0,1]
    a = angle
    a%mean = angle%mean/180
    a%std  = angle%std/180
    ang_dist = init_distrib(a)
    e_dist   = init_distrib(restitution)
    ! Function for ricochet probability
    prob_fnct = init_fnct(rico_prob)       
   
  end subroutine init_ricochet
  subroutine end_ricochet()
  end subroutine end_ricochet

  ! Calculate ricochet for a grain
  ! pos     - position of grain
  ! diam    - diameter of grain
  ! p_in  - impact velocity of grain
  ! p_out - velocity after ricochet of grain
  ! alive   - flag saying whether particle is alive or not
  subroutine grain_ricochet(pos, diam, p_in, p_out, alive)
    type(rvec_t), intent(in)  :: pos      ! position
    real(r_kind), intent(in)  :: diam     ! diameter
    type(rvec_t), intent(in)  :: p_in     ! in momentum
    type(rvec_t), intent(out) :: p_out    ! out momentum
    logical,      intent(out) :: alive    ! did it bounce ?

    select case (decoded_type)
    case (FLAT_T)
       call flat_ricoch(pos, diam, p_in, p_out, alive)
    case (IAN_T)
       call error_msg('Ian Ricoch unimplemented')
       ! call ian_ricoch(pos, diam, p_in, p_out)
    case (THEORETICAL_T)
       call theor_ricoch(pos, diam, p_in, p_out, alive)
    case (FIXED_T)
       call fixed_ricoch(pos, diam, p_in, p_out, alive)
    case (TEIONKEN_T)
       call fixed_ricoch(pos, diam, p_in, p_out, alive)
    case default
       call error_halt(&
            'ricochet.f90: Unknown ricochet function in subroutine ricoch()')
    end select
    ! Kill very slow particles, they contribute nothing and will die anyway
    if (length(p_out)<=grain_min_mom) alive = .false. 
  end subroutine grain_ricochet

  subroutine theor_ricoch(pos, diam, p_in, p_out, alive)
    type(rvec_t),  intent(in)  :: pos    ! position
    real(r_kind), intent(in)   :: diam   ! diameter
    type(rvec_t),  intent(in)  :: p_in   ! in momentum
    type(rvec_t),  intent(out) :: p_out  ! out momentum
    logical, intent(out)       :: alive  ! Did the particle ricochet?
    
    real(r_kind) :: v1  ! in momentum
    real(r_kind) :: a1  ! in angle
    real(r_kind) :: v3  ! out momentum
    real(r_kind) :: a3  ! out angle


    real(r_kind) theta ! impact angle from vertical
    real(r_kind) phi   ! angle on impacting grain
    real(r_kind) alpha ! Angle to grain contact
    real(r_kind) r     ! Effective radius


    real(r_kind) :: a,e 

    real(r_kind) sr ! Surface radius
    real(r_kind) xmin ! minimum perpendicular distance

    real(r_kind) nv,tv,x,y

    ! Now decide if particle bounces or not 
    alive = random_test(eval_fnct(prob_fnct,length(p_in)/grain_mom_scale))
    if (.not. alive)  then
       p_out = null_rvec
       return
    end if 

    a1 = atan2(p_in%y,p_in%x)
    v1 = length(p_in)

    e = random_distrib_num(e_dist)

    a = 1+e ! Normal impulse factor

    sr = diam/2      ! Very rough surface

    r = diam/2 + sr      ! Calulate effective radius

    alpha = asin(sr/r)  ! Calculate angle

    theta = pi/2 + a1   ! Incoming angle is measured anticlockwise
    ! from the horizontal thus for a colliding
    ! particle [-pi,0]

    if (theta>pi/2-alpha) then
       xmin = 1-2*sin(alpha)*cos(theta)
       elseif (theta<pi/2-alpha) then
       xmin = -1
    else
       xmin = sin(theta-alpha)
    end if

    ! x is perpendicular distance from the centre
    !    x = xmin + modulo(xpos/r,2*sin(alpha))*cos(theta)
    call random_number(x)
    x = xmin + 2*sin(alpha)*cos(theta)*x

    y = sqrt(1-x**2)
    phi = -atan2(x,y)

    nv = (a-1)*y  ! Normal speed after collision
    tv = abs(x)-a*friction*y       ! Tangential speed
    if (tv < 0) then
       tv = 0
    else
       tv = sign(tv,x)
    end if

    v3 = v1 * sqrt(nv**2+tv**2)
    a3 = phi - theta + atan2(tv,nv)

    if (a3>pi) then
       a3 = a3-2*pi
    else if (a3<-pi) then
       a3 = a3+2*pi
    end if

    if(a3>pi/2 .or. a3<-pi/2) then
       v3 = e*v3 ! Reduce velocity further
       a3 = a3/2 ! Steepen the velocity
    end if

    a3 = pi/2+a3  ! Convert back to horizontal angle
    call assert(a3>0 .and. a3 < pi,'splash.f90 Ricochet angle error')
    call assert(v3<=v1,'splash.f90 velocity increased !')
    p_out%x = v3*cos(a3)
    p_out%y = v3*sin(a3)
 end subroutine theor_ricoch
 
 ! Ricochet function taken from
 
 ! "Wind modification and bed response during saltation of sand in air"
 ! Acta Mechanica (1991) [Suppl] 1:21--51
 ! R.S. Fixed and P.K. Haff
  subroutine fixed_ricoch(pos, diam, p_in, p_out, alive)
    type(rvec_t), intent(in)  :: pos     ! position
    real(r_kind), intent(in)  :: diam    ! diameter
    type(rvec_t), intent(in)  :: p_in    ! in momentum
    type(rvec_t), intent(out) :: p_out   ! out momentum
    logical, intent(out)      :: alive   ! Did the particle ricochet?

    real(r_kind) :: s 
    real(r_kind) :: ang
    real(r_kind) :: spd

    s = length(p_in)

    ! Now decide if particle bounces or not 
    alive = random_test(eval_fnct(prob_fnct,s/grain_mom_scale))
    if (.not. alive)  then
       p_out = null_rvec
       return
    end if 
   
    ! Calculate random restitution coefficent
    spd = random_distrib_num(e_dist)
    if (spd < 0) then 
       spd = 0
    else if (spd > 1) then 
       spd = 1
    end if

    ! Calculate random angle
    ang = random_distrib_num(ang_dist)
    if (ang < 0)  then
       ang = 0
    else if (ang > pi) then 
       ang = 1
    end if
    ang = ang*pi
    
    p_out = s*spd*rvec_t(cos(ang),sin(ang))

  end subroutine fixed_ricoch

  ! Ricochet function for bouncing off a flat surface with friction
  ! Initial angular velocity assumed to be zero
  ! and speed_mean
  ! Angular momentum not yet accounted for
  subroutine flat_ricoch(pos, diam, p_in, p_out, alive)
    type(rvec_t), intent(in)  :: pos    ! position
    real(r_kind), intent(in)  :: diam   ! diameter
    type(rvec_t), intent(in)  :: p_in   ! in momentum
    type(rvec_t), intent(out) :: p_out  ! out momentum
    logical, intent(out)      :: alive  ! Did the particle ricochet?

    real(r_kind) :: e
    real(r_kind) :: ang
    real(r_kind) :: t_imp
    real(r_kind), parameter :: I = 0.4 ! Moment of inertia ratio for a sphere
    real(r_kind), parameter :: Ir = I/(1+I) ! used in calculation
    
    type(rvec_t) :: p
    type(rmat_t) :: r ! rotation matrix

    ! Now decide if particle bounces or not 
    alive = random_test(eval_fnct(prob_fnct,length(p_in)/grain_mom_scale))
    if (.not. alive)  then
       p_out = null_rvec
       return
    end if

    p = p_in

    do
       ! Calculate a random angle for the surface
       ang = pi*random_distrib_num(ang_dist) 
       r = rotate_matrix(rang_t(ang))
       
       p = r * p   ! Rotate input frame
       e = random_distrib_num(e_dist) ! Calculate a restitution coefficent
       ! Calculation assumes incoming angular velocity is zero
       p_out%y = -e*p%y 
       t_imp = friction*(1+e)*abs(p%y)            ! Tangential impulse
       if (t_imp>Ir*abs(p%x)) t_imp = Ir*abs(p%x) ! Limit slip case
       p_out%x = p%x - sign(t_imp,p%x)
       
       p_out = p_out * r   ! Rotate back, (transposed equation)
       
       if (p_out%y>=0) then ! Did the particle bounce up ?
          exit  ! Yes we're finished
       else            
          p=p_out  ! No calculate another ricochet
       end if
    end do
  end subroutine flat_ricoch

end module ricochet
