

module wind_profile
  use precision
  use sysdep_io
  use maths_const
  use discrete  
  use wind_boundary
  use maths
  use locate
  use track_profile
  use stats
  use wall_law
  use ansi_strings

  implicit none
  private 

  public :: read_wind_profile_nl
  public :: write_wind_profile_nl
  public :: read_wind_state    ! Reads wind velocity
  public :: write_wind_state   ! Write wind velocity
  public :: write_wind_profile ! Writes wind velocity and wind stress 
  public :: calc_wind          ! Time evolves wind velocity
  public :: init_wind          ! Initialize wind state
  public :: end_wind           ! Save wind state
  public :: air_vel            ! Return air velocity in space and time
  public :: get_b_stress       ! Get the wind stress at the ground
  public :: display_wind_step  ! Display info every step
  public :: write_wind_step    ! Write out to data file

  public :: air_viscosity
  public :: air_density

  
  real(r_kind) bot_corr ! Bottom correction for derivative
  real(r_kind) top_corr ! Top correction for derivative
  integer      :: maxit   = 100     ! Max time iterations
  integer      :: minit   = 1       ! Minimum iterations
  real(r_kind) :: eps     = 1e-6    ! Relative error tolerance
  real(r_kind) :: press   = 0       ! Pressure gradient
  real(r_kind) :: yzero   = 1e-5    ! Roughness height
  real(r_kind) :: air_density   = 1.225       ! Density
  real(r_kind) :: air_viscosity = 1.7763e-05  ! Viscosity
  real(r_kind) :: vkarman = 0.41! Von Karman's constant
  real(r_kind) :: lambda      = 1 !1-fully implicit, 0-fully explicit, 0.5-Crank-Nicolson
  logical      :: test_wind    = .false. ! Are we testing this module ?
  logical      :: dynamic_wind = .true.  ! solve time dependent equations
  logical      :: flat_wind    = .false. ! Wind velocity is independent of height
  logical      :: const_wind  = .false. ! Is constant in time
  logical      :: load_state  = .false. ! Is constant in time
  character(len=FILE_NAME_LEN) :: wind_fn = "wind_state.bin" ! Are we load_state ?
  logical      :: three_point = .false. ! Use three point stencil at the ends
  logical      :: zero_slip = .false.  ! Lower boundary condition is u=0

  namelist /wind_profile_nl/ maxit, minit, eps, press, yzero, wind_fn,  &
       air_density, air_viscosity, vkarman, lambda, &
       three_point, zero_slip, &
       test_wind, dynamic_wind, flat_wind, const_wind, load_state

  real(r_kind), allocatable :: stress(:)  ! Air stress 
  real(r_kind), allocatable :: windu(:)   ! Air velocity 
  real(r_kind), allocatable :: mix_len(:) ! Mixing length k^2y^2

  real(r_kind) :: kvisco  ! Kinematic viscosity

  ! Now some other variables for reporting on the behaviour of the module
  type(stats_i_t) :: iter_stats ! How many iterations does convergence take
  type(stats_r_t) :: err_stats  ! Estimate size of max absolute error
contains
  subroutine read_wind_profile_nl(fp)
    integer, intent(in) :: fp
    read(fp, nml=wind_profile_nl)
  end subroutine read_wind_profile_nl
  subroutine write_wind_profile_nl(fp)
    integer, intent(in) :: fp
    write(fp,'(A)')      ' &wind_profile_nl'
    write(fp,'(A,I5)')   '    maxit            = ', maxit
    write(fp,'(A,I5)')   '    minit            = ', minit 
    write(fp,'(A,G11.4)')'    eps              = ', eps
    write(fp,'(A,G11.4)')'    press            = ', press
    write(fp,'(A,G11.4)')'    yzero            = ', yzero 
    write(fp,'(A,G11.4)')'    air_density      = ', air_density
    write(fp,'(A,G11.4)')'    air_viscosity    = ', air_viscosity
    write(fp,'(A,G11.4)')'    vkarman          = ', vkarman
    write(fp,'(A,G11.4)')'    lambda           = ', lambda
    write(fp,'(A,L1)')   '    test_wind        = ', test_wind
    write(fp,'(A,L1)')   '    dynamic_wind     = ', dynamic_wind
    write(fp,'(A,L1)')   '    flat_wind        = ', flat_wind
    write(fp,'(A,L1)')   '    const_wind       = ', const_wind
    write(fp,'(A,L1)')   '    load_state       = ', load_state
    write(fp,'(A,L1)')   '    three_point      = ', three_point
    write(fp,'(A,L1)')   '    zero_slip        = ', zero_slip 
    write(fp,'(A)')      '    wind_fn          = "' // trim(wind_fn) // '"'
    write(fp,'(A)')      ' /'
    write(fp,'(A)')' '
  end subroutine write_wind_profile_nl
  ! Initialize the wind at time t  
  ! Iterate with no particle stress until wind speed converges 
  ! to steady state, for the boundary speed at time 
  subroutine end_wind(dirn)
    character(len=FILE_NAME_LEN), intent(in) :: dirn
    call write_wind_state(dirn)
    deallocate(stress, windu, mix_len)
    call stats_print(std_log, iter_stats,'Wind iterations')
    call stats_print(std_log, err_stats,'Wind errors')
    write(std_log,*)'Parameters'
    write(std_log,'(A,I5)')   '    maxit = ', maxit
    write(std_log,'(A,I5)')   '    minit = ', minit 
    write(std_log,'(A,G11.4)')'    eps   = ', eps
    load_state=.true.
  end subroutine end_wind
  subroutine init_wind(dirn)
    character(len=FILE_NAME_LEN), intent(in) :: dirn
 
    call stats_clear(iter_stats)
    call stats_clear(err_stats)

    kvisco = air_viscosity/air_density

    allocate(stress(ngp), windu(0:ngr+1), mix_len(ngp))
    mix_len = vkarman**2*(grid_y+yzero)**2
    stress = 0
    windu  = 0

    ! Air velocity is defined on cell mid points
    ! These factors are for calculating second order accurate
    ! du/dy at top and bottom of grid
    ! Produces unstable algorithm on top grid point
    if (three_point) then
       bot_corr = 1 + (grid_my(1) -grid_y(1))/(grid_my(2) -grid_y(1))
       top_corr = 1 + (grid_y(ngp)-grid_my(ngr))/(grid_y(ngp)-grid_my(ngr-1))
    else
       top_corr = 1 
       bot_corr = 1 
    end if

    ! This is a very special case used for testing trajectory calculation
    ! The wind is taken to be constant everywhere
    ! The bottom stress is fixed
    if (flat_wind) then
       windu     = get_wind_boundary(real(0,r_kind)) ! Constant wind everywhere
       stress    = 0
       stress(1) = air_density*windu(1)**2       ! Fix the bottom stress
       return
    end if

    ! If we're load_state load the old wind state
    if (load_state) then
       call read_wind_state(dirn) ! Load the old wind velocity
    else
       call calc_init_wind        ! calculate a wind state
    end if
    stress = calc_stress(windu) ! Calculate the stress
  end subroutine init_wind
 
  subroutine calc_init_wind()
    real(r_kind) :: time = 0

    real(r_kind) exactu(0:ngr+1)   ! What the exact wind speed should be
    real(r_kind) oldu(0:ngr+1) ! Wind speed before iteration

    real(r_kind) topb ! Top boundary condition
    real(r_kind) fv ! Friction velocity
    real(r_kind) err(2) ! For absolute and relative errors
    integer n

    ! Calculations assume that lower boundary condition enforces asymptotic
    ! equivalence to law of the wall.
    ! Analytic solution is wrong for zero_slip boundary condition
    ! Otherwise we have to calculate the wind state
    ! First calculate or get  friction velocity
    topb = get_wind_boundary(real(0,r_kind)) ! Constant wind everywhere
    if (boundary_is_stress()) then
       fv = topb
       topb = wall_law_u(grid_y(ngp),yzero,fv,vkarman,kvisco)
    else  ! Solve for friction velocity that gives specified top velocity
       fv = calc_friction_vel(topb,grid_y(ngp),yzero,vkarman,kvisco)
       write(std_out,'(1X,A,F5.3)')'Calculated friction velocity ',fv
    end if

    ! Now calculate exact velocity profile
    exactu = wall_law_u((/grid_y(1),grid_my,grid_y(ngp)/),yzero,fv,vkarman,kvisco)
    call assert(exactu(ngp)-topb,'Friction velocity calc error')
    windu = exactu

    ! Iterate the flow solver just to check it's stable
    do n = 1,5
       oldu = windu
       call calc_wind(time,huge(time))
    end do

    if (test_wind) then
       err = ar_error(windu,oldu) ! How much did it change on the last iteration?
       write(*,'(2(a,es11.4))')&
            ' Change,      absolute ',err(1),', relative ',err(2)
       err = ar_error(windu,exactu) ! How big are the errors?
       write(std_out,'(2(a,es11.4))')&
            ' Wind errors, absolute ',err(1),', relative ',err(2)
       if (err(2)>0.1) then
          write(std_out,'(a,4(G11.4))')' windu  ',windu(0:3)
          write(std_out,'(a,4(G11.4))')' exactu ',exactu(0:3)
          write(std_out,'(a,2(a,(G11.4)))')'top',' windu  ',windu(ngp),&
               ' exactu ',exactu(ngp)
       end if
    end if
  end subroutine calc_init_wind
  subroutine read_wind_state(dirn)
    character(len=*), intent(in) :: dirn
    integer fp
    call assign_lun(fp)
    open(fp,file=trim(dirn)//trim(wind_fn),form='unformatted',&
         status='old',action='read')
    read(fp)windu(1:ngr)
    close(fp);call free_lun(fp)
  end subroutine read_wind_state
  subroutine write_wind_state(dirn)
    character(len=*), intent(in) :: dirn
    integer fp
    call assign_lun(fp)
    open(fp,file=trim(dirn)//trim(wind_fn),form='unformatted',&
         status='replace', action='write')
    write(fp)windu(1:ngr)
    close(fp);call free_lun(fp)
  end subroutine write_wind_state
  subroutine write_wind_profile(fp)
    integer, intent(in) :: fp
    write(fp)windu(1:ngr), stress
  end subroutine write_wind_profile

  ! Calculate the evolution of the wind from time to time+t_step
  subroutine calc_wind(time,t_step)
    use bed
    real(r_kind), intent(in) :: time
    real(r_kind), intent(in) :: t_step

    real(r_kind) norm
    real(r_kind) ap(ngr)   ! Diagonal of tridiagonal system
    real(r_kind) rhs(ngr) 
    real(r_kind) brhs(ngr)
    real(r_kind) dun(ngp)  ! |du/dy| on grid points
    real(r_kind) eviss(ngp)
    real(r_kind) uprev(ngr)!
    real(r_kind) topb      ! Top boundary condition, velocity or friction velocity
    real(r_kind) topu      ! Top boundary velocity

    real(r_kind) tops      ! Top boundary stress (divided by density)
    logical stress_b       ! Top boundary condition is stress
    real(r_kind) err(2)
    real(r_kind) oldu(0:ngr+1) ! Wind speed before iteration
    integer n
    real(r_kind) bot_fv

    oldu = windu
    !...      outer flow velocity:
    ! The top position hold the boundary data
    topb = get_wind_boundary(time) 
    stress_b = boundary_is_stress()
    if (stress_b) then
       tops = topb**2
       topu = 0
    else
       tops = 0
       topu = topb
    end if

    ! Fix everything with a constant wind
    if (flat_wind) then
       windu    = topb ! Constant wind everywhere
       stress   = 0
       stress(1) = air_density*topb**2
       return
    end if

    if (.not. const_wind) then
       !
       !.... press is dp/dx (negative!), get_part_imp=particle source term
       ! A weighted amount of particle stress and the last time steps stress

       ! When we use the static solver the particle force must be averaged
       ! to prevent non-physical oscillations

       brhs = - get_part_force(.not.dynamic_wind)/(air_density*grid_area) !- press/air_density*grid_dy
       brhs(1)=0

      ! Set Right Hand Side for equations
       if (dynamic_wind) then
          brhs = brhs + windu(1:ngr)*grid_dy/t_step 
       end if

       ! Calculate normal velocity gradient dun
       dun  = calc_shear(windu)
       norm  = 1   ! Size of error
       n = 0       ! Iteration count
       ! We iterate the flow solver and average the value of du used
       ! in calculating the viscosity.
       ! A minimum number of iterations is needed to converge to the solution
       ! Due to the nonlinear nature of the problem it seems unstable otherwise
       do    ! Iterate flow solver 
          n=n+1 ! Count the loops
          
          ! Calculate top wind velocity from top_shear stress
          if (stress_b) then
             dun(ngp)=2*tops&
                  /(kvisco+sqrt(kvisco**2+4*tops*mix_len(ngp)))
             topu = windu(ngr) + grid_mdy(ngp)/top_corr*(dun(ngp) &
                  - (1-top_corr)/grid_mdy(ngp-1)*(windu(ngr)-windu(ngr-1)))
          end if
          !bot_fv=sign(sqrt(abs(dun(1))&
!               *(kvisco+mix_len(1)*abs(dun(1)))),dun(1))         
          !windu(0) = wall_law_u(grid_y(1),yzero,bot_fv,vkarman,kvisco)
          if (.true.) then
             bot_fv=calc_friction_vel(windu(1),grid_my(1),yzero,vkarman,kvisco)
             windu(0) = wall_law_u(grid_y(1),yzero,bot_fv,vkarman,kvisco)
          end if
  
          windu(ngp) = topu
          
          ! the big loop
          if (n == maxit .or. (n >= minit .and. norm < eps)) exit 
          
          ! Stabilize iteration scheme for non-linear solver
          dun = (dun+calc_shear(windu))/2

          ! Calculate Effective viscosity
          eviss=(kvisco+mix_len*abs(dun))/grid_mdy
          
          rhs = brhs
          
          if (stress_b) then
             rhs(ngr)   = rhs(ngr)+tops
             eviss(ngp) = 0
          else
             rhs(ngr) = rhs(ngr) + eviss(ngp)*windu(ngr+1)
          end if
          
          ! ground boundary condition
          if (zero_slip) then
             windu(0) = 0
          else
            !bot_fv=calc_friction_vel(windu(1),grid_my(1),yzero,vkarman,kvisco)
   !          bot_fv=sign(sqrt(abs(dun(1))&
   !               *(kvisco+mix_len(1)*abs(dun(1)))),dun(1))
   !          windu(0) = wall_law_u(grid_y(1),yzero,bot_fv,vkarman,kvisco)
   !          if (windu(0)<0) windu(0)=0
  !        rhs(1) = rhs(1) + eviss(1)*windu(0)
          end if
          rhs(1) = rhs(1) + bot_fv**2
          eviss(1) = 0
          
          ap  = eviss(2:ngp) + eviss(1:ngp-1)

          ! Add in time dependent terms
          if (dynamic_wind) then
             ap  = ap + grid_dy/t_step
          end if
 
          ! Calculate the residual vector
          uprev = tridiagmult(windu(1:ngr),-eviss(2:ngr),ap,-eviss(2:ngr))-rhs
          ! Calculate residual error norms
          err = (/maxval(abs(uprev)), sqrt(sum(uprev**2)/ngr)/) 
          
          ! Calculate new wind and see how much it has changed this iteration
          call tridiagsolv(windu(1:ngr),-eviss(2:ngr),ap,-eviss(2:ngr),rhs)

          norm=err(1)               ! Iteration criteria is max absolute error
       end do ! iterate solution loop
       if (n == maxit) then
          write(std_err,"(1X,A,2G12.4,A,10X)")STR_RED//&
               "wind_profile: Max iterations exceeded."&
               //" error ",err,STR_RESET
       end if
       ! Count statistics on the functioning of the algorithm
       call stats_add(iter_stats,n)
       call stats_add(err_stats,norm)
       if(.not.dynamic_wind) windu = (windu+oldu)/2 ! Average wind
     end if

    stress = calc_stress(windu) ! Now calculate the stress on the grid points

  end subroutine calc_wind
  function get_b_stress() result(r)
    real(r_kind) r
    r = stress(1)
  end function get_b_stress
  subroutine display_wind_step(fp)
    integer, intent(in) :: fp
    real(r_kind) az0,bz0,ut,ub,du
    if (stress(1)>0) then
       ub = sqrt(stress(1)/air_density)
    else 
       ub = -sqrt(-stress(1)/air_density)
    end if
    if (stress(ngp)>0) then
       ut = sqrt(stress(ngp)/air_density)
       az0 = (grid_y(ngp) + yzero)* exp(-vkarman*windu(ngp)/ut)
    else
       ut = -sqrt(-stress(ngp)/air_density)
       az0 = 0
    end if
    if (windu(ngr) == windu(ngr-1)) then
       bz0=0
    else
       du = (windu(ngr) + windu(ngr-1))/(2*(windu(ngr) - windu(ngr-1)))
       bz0 = sqrt((grid_my(ngr-1)+ yzero)*(grid_my(ngr)+ yzero)) &
            * ((grid_my(ngr-1)+ yzero)/(grid_my(ngr)+ yzero))**du
    end if
    
    write(fp,'(2(A,F4.2),2(A,F6.2),A,2(ES9.2))')&
         " u*: bottom ", ub ,", top ", ut, &
         " u: bottom ", windu(0) ,", top ", windu(ngp), &
         ", z0 ", az0, bz0
  end subroutine display_wind_step
  subroutine write_wind_step(fp)
    integer, intent(in) :: fp
    real(r_kind) bz0,ut,ub,du
    if (stress(1)>0) then
       ub = sqrt(stress(1)/air_density)
    else 
       ub = -sqrt(-stress(1)/air_density)
    end if
    if (stress(ngp)>0) then
       ut = sqrt(stress(ngp)/air_density)
    else
       ut = -sqrt(-stress(ngp)/air_density)
    end if
     if (windu(ngr) == windu(ngr-1)) then
       bz0=0
    else
       du = (windu(ngr) + windu(ngr-1))/(2*(windu(ngr) - windu(ngr-1)))
       bz0 = sqrt((grid_my(ngr-1)+ yzero)*(grid_my(ngr)+ yzero)) &
            * ((grid_my(ngr-1)+ yzero)/(grid_my(ngr)+ yzero))**du
    end if
    
    write(fp,'(4(F8.3),1(ES9.2))',advance='no')ub,ut,windu(0),windu(ngp),bz0
  end subroutine write_wind_step
  ! Calculate the stress for a particular velocity distribution
  function calc_stress(u) result(s)
    real(r_kind), intent(in) :: u(0:ngp)
    real(r_kind) s(1:ngp)
    real(r_kind) du(1:ngp)
    du=calc_shear(u)
    s=air_density*du*(kvisco+mix_len*abs(du)) 
  end  function calc_stress   
  function calc_shear(u) result(du)
    real(r_kind), intent(in) :: u(0:ngp)
    real(r_kind) du(1:ngp)
    du      = (u(1:ngr+1)-u(0:ngr))/grid_mdy
    du(1)   = bot_corr*du(1)   + (1-bot_corr)*du(2)
    du(ngp) = top_corr*du(ngp) + (1-top_corr)*du(ngp-1)
  end function calc_shear
 ! return the air velocity at time t height y with possible position j
  function air_vel(y,t,j) result(v)
    real(r_kind), intent(in) :: y            ! Current height
    real(r_kind), intent(in) :: t            ! Time
    real(r_kind), intent(in), optional :: j  ! Grid position
    real(r_kind) :: v                        ! Velocity
    integer, save  :: i=1
    if (present(j)) then
       i=j
    else
       i = lookup(y,grid_my,i) ! Mostly we are searching in similar places
    end if
    if (i==0) then
       v = (windu(1)*(y-grid_y(1))+windu(0)*(grid_my(1)-y))/(grid_my(1)-grid_y(1))
    else if (i==ngr) then
       v = (windu(ngr+1)*(y-grid_my(ngr))+windu(ngr)*(grid_y(ngr+1)-y))&
            /(grid_y(ngr+1)-grid_my(ngr))
    else
       v = (windu(i)*(y-grid_my(i))+windu(i+1)*(grid_my(i+1)-y))/grid_mdy(i+1)
    end if
  end function air_vel

   ! Prevent negative velocities but preserve momentum
   subroutine remove_neg(u)
     real(r_kind), intent(inout) :: u(1:ngr)
     integer i
     do i = 1, ngr-1 
        if (u(i)<0) then
           u(i+1) = u(i+1) + u(i)*grid_dy(i)/grid_dy(i+1)
           u(i) = 0
        else 
           exit
        end if
     end do
     if (i>1) write(std_out,*)"negative velocities detected"
     if (u(ngr)<0) then
        write(std_log,*)"negative top velocity detected. dumping windu"
        write(std_log,*)u
        call error_halt("wind_profile.f90: top velocity negative")
     end if
   end subroutine remove_neg
end module wind_profile
  
