module bed
  use precision
  use maths_const
  use sysdep_io
  use vector
  implicit none
  private 
  
  public :: init_surface
!  public :: impact_data
  public :: write_bed_nl
  public :: read_bed_nl
  public :: get_bed_area  ! Return the bed area
  public :: random_bed_pos  ! Return random bed positions
  
  ! Loads and defines contours and particle size distribution

  type bedsurf
     real(r_kind) x  ! x(i)<=x<x(i+1) defines position on surface
     real(r_kind) y  ! 
     real(r_kind) d
  end type bedsurf

  integer :: nix
  real(r_kind) hmax
  real(r_kind) :: bed_area=-1
  type(bedsurf), allocatable  :: sur(:)

  ! Input variables
  real(r_kind) :: bed_length = 1        ! Bed is ten meters long
  real(r_kind) :: bed_width  = 0.0001   !0.1mm wide
  logical      :: bed_file   = .false.  ! Read heights from a file ?

  namelist /bed_nl/ bed_length, bed_width, bed_file
contains
  subroutine read_bed_nl(fp)
    integer, intent(in) :: fp
    integer iostat
    read(fp, nml=bed_nl,err=1,iostat=iostat)
    return
1   call print_iostat(iostat,std_err,'Error with bed_nl reading file "'//trim(get_name(fp))//'"')
    call error_halt
  end subroutine read_bed_nl
  subroutine write_bed_nl(fp)
    integer, intent(in) :: fp
    write(fp,'(A)')      ' &bed_nl'
    write(fp,'(A,G11.4)')'    bed_length = ', bed_length 
    write(fp,'(A,G11.4)')'    bed_width  = ', bed_width 
    write(fp,'(A,L1)')   '    bed_file  = ', bed_file
    write(fp,'(A)')      ' /'
    write(fp,'(A)')' '
  end subroutine write_bed_nl
  subroutine init_surface(dirn)
    character(len=*), intent(in) :: dirn
    integer fp
    integer iostat
    bed_area=bed_length*bed_width 
    if (bed_file) then
       call assign_lun(fp)
       write(std_out,*)'Reading surface data from ' // trim(dirn) // 'surface'
       call err_open(fp, file=trim(dirn) // 'surface', status='old')
       read(fp,*,err=1,iostat=iostat)nix
       allocate(sur(1:nix))
       read(fp,*,err=1,iostat=iostat)sur
       write(std_out,*)'Surface contains', nix, 'points'
       call err_close(fp); 
    else
       write(std_out,*)'Surface is flat'
       nix = 2
       allocate(sur(1:nix))
       sur(1)%x = 0
       sur(2)%x = bed_length
       sur%y    = 0
    end if
    hmax=maxval(sur%y)
    return
1   call print_iostat(iostat,std_err,'Error reading file "'//trim(get_name(fp))//'"')
    call error_halt
  end subroutine init_surface
  function get_surface(x) result(s)
    real(r_kind), intent(in) :: x
    type(bedsurf) s
    integer i
    if (.not. bed_file) then
       s = bedsurf(0,0,0)
    else
       
       ! Find where in the surface it impacts
       if (x < sur(1)%x) then
          i=nix
       else if (x >= sur(nix)%x) then
          i=1
       else
          i=0     !   find impact point's position
          do
             if (sur(i+1)%x > x) exit
             i=i+1
          enddo
       endif
       s = sur(i)
    end if
  end function get_surface  
  subroutine random_bed_pos(r)
    type(rvec_t), intent(out) :: r(:)
    real(r_kind) tmp_rand(size(r))
    call random_number(tmp_rand)
    r%x=bed_length*tmp_rand
    if (.not. bed_file) then
       r%y=0
    else
       r%y=0
       write(std_err,*)'bed.f90: Unimplemented random_bed_position'
    end if
  end subroutine random_bed_pos
  function get_bed_area()
    real(r_kind) get_bed_area
    call assert(bed_area>0,'bed.f90:Bed area requested before calculated')
    get_bed_area = bed_area
  end function get_bed_area
  ! Caclulates exact colliision position for a particle 
  ! and returns this position and diameter of particle with which
  ! collision occurs
  subroutine impact_data(rc,dc,ri,di,vi)
    type(rvec_t), intent(out) :: rc ! Collision position
    real(r_kind), intent(out) :: dc ! Diameter of colliding particle
    type(rvec_t), intent(in)  :: ri ! Collision position
    real(r_kind), intent(in)  :: di ! Diameter of colliding particle
    type(rvec_t), intent(in)  :: vi ! velocity of incoming particle
    
    real(r_kind) xs,xn1,dh,xn,zn,znn,x0,zn1
    integer ix, numx 

    real(r_kind) vr,tanv,sinv,cosv 

    vr = length(vi)
    tanv = vi%y/vi%x
    sinv = vi%y/vr
    cosv = vi%x/vr

    xs=ri%x+vi%y*hmax/vi%x   ! What is bottom possible collision spot 

    if (ri%x < sur(1)%x) then
       ix=nix
    else if (ri%x >= sur(nix)%x) then
       ix=1
    else
       ix=0     !   find impact point's position
       do
          if (sur(ix+1)%x > ri%x) exit
          ix=ix+1
       enddo

    endif
    !
    !...  this takes xn1 to be x pos of lowest point of incoming grain
    xn1=sur(ix)%x+di*sinv
    if (xn1 < 0.0) xn1=xn1+bed_length
    dh=-(ri%x-sur(ix)%x)*tanv
    if (ix == nix) dh=-(ri%x+bed_length-sur(ix)%x)*tanv
    zn1=hmax+dh-di*cosv

    do
       ix=ix+1
       if (ix > nix) then
          ix=1
          xn1=xn1-bed_length
       endif
       xn=sur(ix)%x-sur(ix)%d*sinv
       if (xn > bed_length) xn=xn-bed_length
       zn=sur(ix)%y+sur(ix)%d*cosv
       !....   this takes initial projection point to lowest part of a grain
       znn=zn1+(xn-xn1)*tanv
       if (znn < zn) exit
    end do

    numx=ix
    !
    !---- exact coll point
    !
    rc%x = sur(ix)%x-(xn1-di*sinv)
    if (x0 < 0.0) x0=x0+bed_length
    rc%y = sur(ix)%y
    dc=sur(ix)%d
    end subroutine impact_data
end module bed
