! Module contains variables describing the grid
! All constant after the program has read in the initial data
module discrete
  use precision
  use sysdep_io
  use bed
  use locate
  implicit none
  private

  public :: init_grid
  public :: write_grid
  public :: check_coord
  public :: adjust_coord
  public :: read_grid_nl
  public :: write_grid_nl

  integer, public :: ngp=-1 ! Number of grid points
  integer, public :: ngr=-1 ! Number of grid regions ngp-1

  ! Should be a derived type but since debugger is broken...
  real(r_kind), allocatable, public :: grid_y(:)    ! Grid spaces
  real(r_kind), allocatable, public :: grid_dy(:)   ! Difference
  real(r_kind), allocatable, public :: grid_mdy(:)  ! Mid point difference
  real(r_kind), allocatable, public :: grid_my(:)   ! Mid point
  real(r_kind), allocatable, public :: grid_vol(:)  ! Volume

  real(r_kind), public :: grid_area=-1  ! Area of grid

  ! Note that parameters for calculating grid use double precision
  real(rd_kind) :: lambda

  integer       :: grid_ny     = 100     ! Number of vertical grid points
  real(rd_kind),public :: grid_height = 2       ! Highest grid point
  real(rd_kind) :: grid_scale  = 0.00003 ! Offset length scale

  namelist/grid_nl/ grid_ny, grid_height, grid_scale

  character(len=*), parameter :: grid_fn = 'grid.bin'

contains
  subroutine read_grid_nl(fp)
    integer, intent(in) :: fp
    integer iostat
read(fp, nml=grid_nl,err=1,iostat=iostat)
return
1 call print_iostat(iostat,std_err,'Error with grid_nl reading file "'//trim(get_name(fp))//'"')
  call error_halt
  end subroutine read_grid_nl
  subroutine write_grid_nl(fp)
    integer, intent(in) :: fp
    write(fp,'(A)')      ' &grid_nl'
    write(fp,'(A,I5)')   '    grid_ny     = ', grid_ny
    write(fp,'(A,G11.4)')'    grid_height = ', grid_height
    write(fp,'(A,G11.4)')'    grid_scale  = ', grid_scale
    write(fp,'(A)')      ' /'
    write(fp,'(A)')' '
  end subroutine write_grid_nl

  ! We can read the grid in from a file if we prefer
  subroutine read_grid(dirn)
    character(len=*), intent(in) :: dirn
    integer fp
    integer iostat
    write(std_out,'(1x,a)') 'Reading the grid spacing from "' &
         // trim(dirn) // grid_fn // '"'
    call assign_lun(fp)
    call err_open(fp,file=trim(dirn) // grid_fn, status='old')
    read(fp,*,err=1,iostat=iostat) ngp
    ngr = ngp-1
    allocate(grid_y(ngp), grid_dy(1:ngr))
    read(fp,*,err=1,iostat=iostat)grid_y
    call err_close(fp);  
    write(std_out,*) 'Grid file read.'
    grid_dy(1:ngr) = grid_y(2:ngp)-grid_y(1:ngp-1)
    return
1   call print_iostat(iostat,std_err,'Error reading file "'//trim(get_name(fp))//'"')
    call error_halt 
  end subroutine read_grid

   ! Write the grid to a file in the same format
  subroutine write_grid(dirn)
    character(len=*), intent(in) :: dirn
    integer fp
    write(std_out,'(a)') ' Writing the grid to "' // trim(dirn) // grid_fn // '"'
    call assign_lun(fp)
    call err_open(fp,file=trim(dirn) // grid_fn, status='replace')
    write(fp,*) ngp
    write(fp,*) grid_y
    call err_close(fp);  
  end subroutine write_grid
  subroutine deallocate_grid()
    deallocate(grid_y, grid_dy, grid_vol, grid_my, grid_mdy)
  end subroutine deallocate_grid
  subroutine init_grid(dirn)
    character(len=*), intent(in) :: dirn
    integer k
    lambda = (grid_height/grid_scale+1)**(1/real(grid_ny-1,rd_kind))
    ngp = grid_ny
    ngr = grid_ny-1
    allocate(grid_y(ngp), grid_dy(ngr), &
         grid_vol(ngr), grid_my(ngr), grid_mdy(ngp))

    grid_y  = grid_scale*(lambda**(/(k,k=0,ngp-1)/)-1)
    grid_dy = grid_scale* lambda**(/(k,k=0,ngr-1)/)*(lambda-1)
    grid_area = get_bed_area()
    grid_vol = grid_area*grid_dy(1:ngr)
!    grid_my(0)     = grid_y(1)
    grid_my(1:ngr) = (grid_y(2:ngp)+grid_y(1:ngp-1))/2
!    grid_my(ngp)   = grid_y(ngp)
    grid_mdy(1)       = grid_dy(1)/2
    grid_mdy(2:ngp-1) = (grid_dy(2:ngr)+grid_dy(1:ngr-2))/2
    grid_mdy(ngp)     = grid_dy(ngr)/2

    call assert(all(grid_dy>=0),'discrete.f90 some grid spacing is negative')
    call assert(grid_dy(1:ngr)+grid_y(1:ngp-1)-grid_y(2:ngp), &
         "Miscalculated grid_dy")  
    call write_grid(dirn)
  end subroutine init_grid    
 
 ! Interpolates function  defined at grid points
  ! g(y) = f*g(coord)+(1-f)*g(coord+1)
  subroutine grid_interp(y,f,coord)
    real(r_kind), intent(in) :: y
    real(r_kind), intent(out) :: f
    integer, intent(in), optional :: coord
    integer i
    if (present(coord)) then
       i = coord
    else
       i = grid_coord(y)
    end if
    f = (grid_y(i+1)-y)/grid_dy(i);
    call assert(f>=0 .and. f <=1,'discrete.f90: Error in interp(y,f,i)')
  end subroutine grid_interp

  ! Interpolates function  defined in grid regions
  ! g(y) = f*g(coord)+(1-f)*g(coord+1)
  ! i is s.t. grid_y(i)<y<=grid_y(i+1)
  ! but after the call
  ! i is s.t. grid_my(i)<y<=grid_my(i+1)
  subroutine grid_my_interp(y,f,i)
    real(r_kind), intent(in) :: y
    real(r_kind), intent(out) :: f
    integer, intent(inout) :: i
    if (y<grid_my(1)) then       ! Bottom value is special
       i = 0
       f = (grid_my(1)-y)/(grid_my(1)-grid_y(1))
    else 
       if (y<grid_my(i)) i = i-1
       f = (grid_my(i+1)-y)/(grid_my(i+1)-grid_my(i))
    end if
    call assert(f>=0 .and. f <=1,'discrete.f90: Error in grid_my_interp(y,f,i)')
  end subroutine grid_my_interp

  ! return i s.t. grid_y(i)<y<=grid_y(i+1)
  ! Top grid spacing are smaller so step down
  function grid_coord(y) result(i)
    real(r_kind), intent(in) :: y
    integer i
    do i=size(grid_y)-1,1,-1
       if (y>=grid_y(i)) exit
    end do
  end function grid_coord

  function check_coord(y,i)
    real(r_kind), intent(in) :: y
    integer, intent(in) :: i
    logical check_coord
    check_coord = y >= grid_y(i) .and. y < grid_y(i+1)
  end function check_coord  
  ! Return new grid coordinate if only changed by one
  function adjust_coord(y,old) result (new)
    real(r_kind), intent(in) :: y
    integer, intent(in) :: old
    integer :: new
    if (y >= grid_y(old+1)) then
       new = old+1
       call assert(y<grid_y(new+1),&
            'discrete.f90 y too large in adjust_coord') 
    else if (y >= grid_y(old)) then
       new = old
    else
       new = old-1
       call assert(y>=grid_y(new),&
            'discrete.f90 y too small in adjust_coord') 
    end if
  end function adjust_coord

end module discrete
