module entrainment
  use precision	 
  use particle
  use traj_samp
  use bed
  use maths
  use random
  use grain_var
  use sysdep_io
  use random_distrib

  implicit none
  private
  public :: aero_entran
  public :: read_entrainment_nl
  public :: write_entrainment_nl
  public :: init_entrainment
  public :: end_entrainment

  ! derived parameters for each grain class
  real(r_kind), allocatable :: threshold(:)
  real(r_kind), allocatable :: entran_factor(:) 

  ! Namelist parameters
  real(r_kind) :: tau_threshold = 0.1 ! From Bagnold
  real(r_kind) :: entran_fac    = 1e-4

  type(distrib_pub_t) :: x_vel = distrib_pub_t("normal",3,0.1)
  type(distrib_pub_t) :: y_vel = distrib_pub_t("normal",3,0.1)

  namelist /entrainment_nl/ tau_threshold, entran_fac, x_vel, y_vel

  type(distrib_t) :: vel_x_dist = null_distrib_t
  type(distrib_t) :: vel_y_dist = null_distrib_t

contains
  subroutine init_entrainment()
    allocate(entran_factor(grain_size_num),threshold(grain_size_num))
    entran_factor = get_bed_area()*entran_fac*grain_prob/grain_mom_scale
    threshold     = tau_threshold**2*grain_vel_scale**2*grain_dens_scale
    vel_x_dist = init_distrib(x_vel)
    vel_y_dist = init_distrib(y_vel)
  end subroutine init_entrainment 
  subroutine end_entrainment()
    deallocate(entran_factor,threshold)
  end subroutine end_entrainment
  subroutine read_entrainment_nl(fp) 
    integer, intent(in) :: fp 
    integer iostat
    read(fp,nml=entrainment_nl,err=1,iostat=iostat)
    return
1   call print_iostat(iostat,std_err,&
         'Error with entrainment_nl reading file "'//trim(get_name(fp))//'"')
    call error_halt 
  end subroutine read_entrainment_nl
  subroutine write_entrainment_nl(fp)
    integer, intent(in) :: fp
    write(fp,'(A)')       ' &entrainment_nl'
    write(fp,'(A,G11.4)')'    tau_threshold = ', tau_threshold
    write(fp,'(A,G11.4)')'    entran_fac    = ', entran_fac
    write(fp,'(A,G11.4)',advance='no')'    x_vel         = '
    call write_distrib(fp,x_vel)
    write(fp,'(A,G11.4)',advance='no')'    y_vel         = '
    call write_distrib(fp,y_vel)
    write(fp,'(A)')       ' /'
    write(fp,'(A)')' '
  end subroutine write_entrainment_nl
  ! Given a surface stress tau calculate the number of grains 
  ! entrained in t seconds
  ! Does not entrain grains for negative stress
  subroutine aero_entran(tau, t, dt, aero, g)
    real(r_kind),intent(in)      :: tau       ! Surface air stress
    real(r_kind),intent(in)      :: t         ! Current time
    real(r_kind),intent(in)      :: dt        ! Time interval to cover
    integer,intent(out)          :: aero      ! Number of new aero grains
    type(particles), intent(out) :: g(:)      ! Space for new grains
    real(r_kind) myu(grain_size_num)
    integer n(grain_size_num)
    integer i,k 

    ! Calculate mean numbers in each class
    myu = dt*entran_factor*(tau-threshold)
    where (myu<0) myu = 0
    ! Calculate number of grains in each size class
    do i=1,grain_size_num
       n(i) = random_poisson(myu(i))
    end do
    aero = sum(n)

    ! Limit number of grains in a reasonable way
    if (aero>size(g)) then
       write(std_out,*)'entrainment.f90: Limiting grain numbers'
       n = floor(n*real(size(g))/real(aero))
       aero = sum(n)
    end if

    ! Set class for each grain
    k=0
    do i=1,size(n)
       g(k+1:k+n(i))%gc=i
       k=k+n(i)
    end do

    ! Grain positions are chosen uniformly on the bed at height 0
    ! Grain times are chosen uniformly on the interval (t,t+dt)
    ! There are no parent grains
    ! The diameters are chosen but the function grain_diam
    ! Velocities are chosen from a cartesian distribution

    g(1:aero)%origin = ORIGIN_AERO    
    call random_bed_pos(g(1:aero)%pos)
    g(1:aero)%gy=1
    g(1:aero)%diam   = grain_diam(g(1:aero)%gc) 
    g(1:aero)%time   = t+dt*random_uniform(aero)
    g(1:aero)%parent = 0                
    g(1:aero)%vel%x  = grain_vel_scale*random_distrib_num(vel_x_dist,aero)
    g(1:aero)%vel%y  = grain_vel_scale*random_distrib_num(vel_y_dist,aero)
    call add_grain(g(1:aero))           
  end subroutine aero_entran
end module entrainment
