module impact_stats
  use sysdep_io
  use precision
  use discrete
  use grain_var
  implicit none  
  private

  public :: init_impact_stats    ! Called at start of program
  public :: end_impact_stats     ! Called at end of program
  public :: start_impact_stats   ! Called at start of time step
  public :: finish_impact_stats  ! Called at start of time step
  public :: add_impact_stats     ! Add on statistics for another impact
  public :: add_aero_total_stats ! Add on statistics for aerodynamic and total
  public :: disp_impact_stats    ! Display the statistics for one time step
  public :: disp_t_impact_stats  ! Display the total statistics
  public :: get_v_fluxes         ! Return vertical fluxes
  public :: disp_v_fluxes        ! Display vertical fluxes
  public :: write_v_fluxes        ! Display vertical fluxes
  public :: get_tv_fluxes        ! Return total vertical fluxes
  public :: disp_tv_fluxes       ! Display total vertical fluxes
  public :: write_impact_state   ! Write state
  public :: read_impact_state    ! Read the impact state

  ! Variables and functions for keeping track of statistics for the
  ! number of particles, ejected particles etc...
 
  integer :: died     ! Number of dead particles
  integer :: ejecta   ! Number of ejected particles
  integer :: impacts  ! Number of impacts
  integer :: ricoch   ! Number of bounced particles
  integer :: aero     ! Num of aerodynamically entrained particles
  integer :: total    ! Total number of particles

  integer :: t_died     ! Number of dead particles
  integer :: t_ejecta   ! Number of ejected particles
  integer :: t_impacts  ! Number of impacts
  integer :: t_ricoch   ! Number of bouncedi particles
  integer :: t_aero     ! Num of aerodynamically entrained particles
  integer :: t_total    ! Total number of particles

  real(r_kind) elap_time
  real(r_kind) step_time

contains
  ! Called at end of program
  subroutine end_impact_stats()
    call disp_t_impact_stats(std_out) 
    call disp_tv_fluxes(std_out)
    write(std_out,*)
  end subroutine end_impact_stats
  ! Called at start of program
  subroutine init_impact_stats()
    call start_impact_stats(.true.)
    step_time = huge(step_time)
  end subroutine init_impact_stats
  ! Prepare grain stats for this iteration
  ! Add on current values to totals
  subroutine start_impact_stats(reset)
    logical, intent(in) :: reset      ! Whether or not to reset counters
    if (reset .and. .false.) then ! reset totals
       elap_time = 0
       t_total = 0
       t_aero = 0
       t_died = 0
       t_ejecta = 0
       t_impacts = 0
       t_ricoch = 0
    end if
    died=0
    ejecta=0
    impacts=0 
    ricoch=0;
    aero = 0 
    total = 0
  end subroutine start_impact_stats
  subroutine finish_impact_stats(dt)
    real(r_kind) :: dt ! How much time elapsed this step
    ! Add on totals
    t_total = t_total+total
    t_aero = t_aero+aero
    t_died = t_died + died
    t_ejecta = t_ejecta + ejecta
    t_impacts = t_impacts + impacts
    t_ricoch = t_ricoch + ricoch
    elap_time=elap_time+dt
    step_time=dt
  end subroutine finish_impact_stats
    
  subroutine add_aero_total_stats(a,t)
    integer, intent(in):: a,t
    aero = aero+a
    total = total+t
  end subroutine add_aero_total_stats
      
  subroutine add_impact_stats(ejected,alive)
    integer, intent(in):: ejected ! Number of new grains ejected
    logical, intent(in):: alive ! Whether the mother grain is alive
    if(alive) then
       ricoch = ricoch + 1
    else 
       died = died + 1
    end if
    impacts = impacts + 1
    ejecta = ejecta + ejected
  end subroutine add_impact_stats
  subroutine disp_impact_stats(fp)
    integer, intent(in) :: fp
    write(fp,*)'total',total,'impacts', impacts, &
         'ricoh',ricoch,'died', died,&
         'ejected', ejecta, 'aero', aero
  end subroutine disp_impact_stats
  subroutine disp_t_impact_stats(fp)
    integer, intent(in) :: fp
    write(fp,'(1X,A19,3(A11,I7))')'Total impact stats:',&
         '  total   =', t_total,&
         '  impacts =', t_impacts,&
         '  ricoh   =', t_ricoch
    write(fp,'(20X,3(A11,I7))')&
         '  died    =', t_died,&
         '  ejected =', t_ejecta,&
         '  aero    =', t_aero
  end subroutine disp_t_impact_stats
  subroutine disp_v_fluxes(fp)
    integer, intent(in) :: fp
    real(r_kind) a
    if(step_time==0) then
       a = 0
    else
       a = grain_mass_scale/(step_time*grid_area)
    end if
    write(fp,'(2(A,F8.4))',advance='no')&
         ' down ',a*impacts,&
         ' up ',a*(ricoch+ejecta+aero)
  end subroutine disp_v_fluxes
  subroutine write_v_fluxes(fp)
    integer, intent(in) :: fp
    real(r_kind) a
    if(step_time==0) then
       a = 0
    else
       a = grain_mass_scale/(step_time*grid_area)
    end if
    write(fp,'(2(F8.4))',advance='no')a*impacts,a*(ricoch+ejecta+aero)
  end subroutine write_v_fluxes
  function get_v_fluxes()
    real(r_kind) get_v_fluxes(2)
    real(r_kind) a
    if(step_time==0) then
       a = 0
    else
       a = grain_mass_scale/(step_time*grid_area)
    end if
    get_v_fluxes =a*(/impacts,ricoch+ejecta+aero/)
  end function get_v_fluxes
  subroutine disp_tv_fluxes(fp)
    integer, intent(in) :: fp
    real(r_kind) a
    a = grain_mass_scale/(elap_time*grid_area)
    write(fp,'(2(A,F8.4))',advance='no')&
         ' down ',a*t_impacts,&
         ' up ',a*(t_ricoch+t_ejecta+t_aero)
  end subroutine disp_tv_fluxes
  function get_tv_fluxes()
    real(r_kind) get_tv_fluxes(2)
    real(r_kind) a
    a = grain_mass_scale/(elap_time*grid_area)
   get_tv_fluxes =a*(/t_impacts,t_ricoch+t_ejecta+t_aero/)/(elap_time*grid_area)
  end function get_tv_fluxes
  ! The trace has finished Normalize everything
  subroutine write_impact_state(fp)
    integer, intent(in) :: fp
    write(fp)elap_time,t_total,t_impacts,t_ricoch,t_ejecta,t_aero
  end subroutine write_impact_state
  subroutine read_impact_state(fp)
    integer, intent(in) :: fp
    integer iostat
    read(fp,err=1,iostat=iostat)&
       elap_time,t_total,t_impacts,t_ricoch,t_ejecta,t_aero  
    return
1   call print_iostat(iostat,std_err,'Error reading file "'//trim(get_name(fp))//'"')
    call error_halt
  end subroutine read_impact_state
end module impact_stats
