! Module deals with particle ricochets
! Contains functions for reading and writing parameter values
! And master ricochet function that can choose any number of sub functions

module ricochet
  use sysdep_io
  use precision
  use maths_const
  use grain_var
  use random
  use vector
  use string_param
  use random_distrib
  use maths_fnct
  private

  public :: read_ricochet_nl    
  public :: write_ricochet_nl    
  public :: init_ricochet    
  public :: end_ricochet    
  public :: grain_ricochet       ! Calculate the results of a ricochet

  integer, parameter :: UNINIT_T      = -1
  integer, parameter :: NONE_T        = 1
  integer, parameter :: FLAT_T        = 2
  integer, parameter :: MCEWAN_T      = 3
  integer, parameter :: THEORETICAL_T = 4
  integer, parameter :: FIXED_T       = 5
  integer, parameter :: TEIONKEN_T    = 6
  integer :: decoded_type = UNINIT_T  ! Decode the string for the function type

  ! flat - like bouncing of a flat plate with 
  !        collision determined by restitution distribution, friction
  !        ejection angle gets offset by angle distribution
  !
  ! fixed - ejection is independent of impact angle
  !       - restitution and angle determined by fixed distribution 
        
  ! List of known function types
  type(param_table_t), parameter :: types(6) = (/&
       param_table_t("none",       NONE_T),&  
       param_table_t("flat",       FLAT_T),&  
       param_table_t("mcewan",     MCEWAN_T),&
       param_table_t("theoretical",THEORETICAL_T),&
       param_table_t("teionken",   TEIONKEN_T),&
       param_table_t("fixed",      FIXED_T)/)

  ! string describing type of ricochet function
  character(len=STRING_LENGTH) :: ricochet_type='fixed' 
  real(r_kind) :: friction    = .3    ! contact friction
  type(fnct_pub_t)    :: rico_prob= fnct_pub_t("exponential",(/0.,.94,0.1,0./))
  type(distrib_pub_t) :: angle        = distrib_pub_t("normal",38,10)
  type(distrib_pub_t) :: restitution  = distrib_pub_t("normal",0.56,0.2)
  real(r_kind) :: roughness = 1

  namelist /ricochet_nl/ ricochet_type, friction, rico_prob, &
       angle, restitution, roughness

  type(fnct_t)    :: prob_fnct ! Function for probability of ricochet
  type(distrib_t) :: ang_dist  ! Angle distribution
  type(distrib_t) :: e_dist    ! Restitution distribution
 
contains
  subroutine read_ricochet_nl(fp)
    integer, intent(in) :: fp
    integer iostat
read(fp,nml=ricochet_nl,err=1,iostat=iostat)
return
1 call print_iostat(iostat,std_err,&
       'Error with ricochet_nl reading file "'//trim(get_name(fp))//'"')
  call error_halt
  end subroutine read_ricochet_nl
  subroutine write_ricochet_nl(fp)
    integer, intent(in) :: fp  
    write(fp,'(A)')      ' &ricochet_nl'
    write(fp,'(A)')      '    ricochet_type = "' // trim(ricochet_type) // '"'
    write(fp,'(A,G11.4)')'    friction      = ', friction
    write(fp,'(A,G11.4)')'    roughness     = ', roughness
    write(fp,'(A,G11.4)',advance='no')'    rico_prob     = '
    call write_fnct(fp,rico_prob)
    write(fp,'(A,G11.4)',advance='no')'    angle         = '
    call write_distrib(fp,angle)
    write(fp,'(A,G11.4)',advance='no')'    restitution   = '
    call write_distrib(fp,restitution)
    write(fp,'(A)')      ' /'
    write(fp,'(A)')' '
  end subroutine write_ricochet_nl
  subroutine init_ricochet()
    type(distrib_pub_t) a
    decoded_type = decode_string_param(ricochet_type,types)

    ! Convert angles to a scale of [0,1]
    a = angle
    a%mean = angle%mean/180
    a%std  = angle%std/180
    ang_dist = init_distrib(a)
    e_dist   = init_distrib(restitution)
    ! Function for ricochet probability
    prob_fnct = init_fnct(rico_prob)       
   
  end subroutine init_ricochet
  subroutine end_ricochet()
  end subroutine end_ricochet

  ! Calculate ricochet for a grain
  ! pos     - position of grain
  ! diam    - diameter of grain
  ! p_in  - impact velocity of grain
  ! p_out - velocity after ricochet of grain
  ! alive   - flag saying whether particle is alive or not
  subroutine grain_ricochet(pos, diam, p_in, p_out, alive)
    type(rvec_t), intent(in)  :: pos      ! position
    real(r_kind), intent(in)  :: diam     ! diameter
    type(rvec_t), intent(in)  :: p_in     ! in momentum
    type(rvec_t), intent(out) :: p_out    ! out momentum
    logical,      intent(out) :: alive    ! did it bounce ?

    select case (decoded_type)
    case (UNINIT_T)
       call error_halt(&
            'ricochet.f90: Uninitialized ricochet type in grain_ricochet')
    case (NONE_T)
       alive = .false.
       p_out = null_rvec
    case (FLAT_T)
       call flat_ricoch(pos, diam, p_in, p_out, alive)
    case (MCEWAN_T)
       call error_msg('Mcewan Ricoch unimplemented')
       ! call mcewan_ricoch(pos, diam, p_in, p_out)
    case (THEORETICAL_T)
       call theor_ricoch(pos, diam, p_in, p_out, alive)
    case (FIXED_T)
       call fixed_ricoch(pos, diam, p_in, p_out, alive)
    case (TEIONKEN_T)
       call fixed_ricoch(pos, diam, p_in, p_out, alive)
    case default
       call error_halt(&
            'ricochet.f90: Unknown ricochet function in subroutine ricoch()')
    end select
    ! Kill very slow particles, they contribute nothing and will die anyway
    if (length(p_out)<=grain_min_mom) alive = .false. 
  end subroutine grain_ricochet

  subroutine theor_ricoch(pos, diam, p_in, p_out, alive)
    type(rvec_t),  intent(in)  :: pos    ! position
    real(r_kind), intent(in)   :: diam   ! diameter
    type(rvec_t),  intent(in)  :: p_in   ! in momentum
    type(rvec_t),  intent(out) :: p_out  ! out momentum
    logical, intent(out)       :: alive  ! Did the particle ricochet?
    
    real(r_kind) :: p_len  ! in momentum

    real(r_kind) cos_alpha ! Angle to grain contact
    real(r_kind) sin_alpha ! Angle to grain contact
    real(r_kind) r     ! Effective radius


    real(r_kind) :: e 

    real(r_kind) sr ! Surface radius
    real(r_kind) xmin ! minimum impact parameter
    real(r_kind) xmax ! maximum impact parameter

    real(r_kind) nv,tv

    real(r_kind) sin_theta, cos_theta
    real(r_kind) t_imp
    real(r_kind), parameter :: I = 0.4 ! Moment of inertia ratio for a sphere
    real(r_kind), parameter :: Ir = I/(1+I) ! used in calculation

    type(rvec_t) p
 
    ! Now decide if particle bounces or not 
    alive = random_test(eval_fnct(prob_fnct,length(p_in)/grain_mom_scale))
    if (.not. alive)  then
       p_out = null_rvec
       return
    end if 
    
    cos_alpha = 1-roughness           !    ! Surface critical angle
    sin_alpha = sqrt(1-cos_alpha**2)

    p=p_in
    
    do   ! Keep bouncing until it bounces up
 
       p_len = length(p)
       sin_theta = -p%y/p_len
       cos_theta = p%x/p_len
    
       e = random_distrib_num(e_dist)
       
       
       if (sin_theta<sin_alpha) then
          if (cos_theta<0) then
             xmin = -1
             xmax = -1 + 2*sin_alpha*sin_theta
          else
             xmin = 1-2*sin_alpha*sin_theta
             xmax = 1
          end if
       else
          xmin = cos_theta*cos_alpha-sin_alpha*sin_theta
          xmax = cos_theta*cos_alpha+sin_alpha*sin_theta
       end if
       
       ! tv  is impact parameter - tangent unit vector component
!!$       tv = random_uniform() ! This is very good for debugging
!!$       if (tv<.3) then
!!$          tv = xmin
!!$       else if (tv>0.7) then
!!$          tv = xmax
!!$       else
!!$          tv = (xmin+xmax)/2
!!$       end if
       tv = xmin + (xmax-xmin)*random_uniform()
       nv = sqrt(1-tv**2)  ! Normal unit vector
       p_out%y = -e*nv
       t_imp = friction*(1+e)*nv            ! Tangential impulse
       if (t_imp>Ir*abs(tv)) t_imp = Ir*abs(tv) ! Limit slip case
       p_out%x = tv - sign(t_imp,tv)
       
       ! Rotate back
       p_out=p_len&
            *rmat_t(rvec_t(sin_theta,cos_theta),rvec_t(-cos_theta,sin_theta))&
            *(rmat_t(rvec_t(nv,-tv),rvec_t(-tv,-nv))*p_out)
       
       if (p_out%y>=0) then ! Did the particle bounce up ?
          exit  ! Yes we're finished
       else            
          p=p_out  ! No calculate another ricochet
       end if
    end do
       
  end subroutine theor_ricoch
 
 ! Ricochet function taken from
 
 ! "Wind modification and bed response during saltation of sand in air"
 ! Acta Mechanica (1991) [Suppl] 1:21--51
 ! R.S. Fixed and P.K. Haff
  subroutine fixed_ricoch(pos, diam, p_in, p_out, alive)
    type(rvec_t), intent(in)  :: pos     ! position
    real(r_kind), intent(in)  :: diam    ! diameter
    type(rvec_t), intent(in)  :: p_in    ! in momentum
    type(rvec_t), intent(out) :: p_out   ! out momentum
    logical, intent(out)      :: alive   ! Did the particle ricochet?

    real(r_kind) :: s 
    real(r_kind) :: ang
    real(r_kind) :: spd

    s = length(p_in)

    ! Now decide if particle bounces or not 
    alive = random_test(eval_fnct(prob_fnct,s/grain_mom_scale))
    if (.not. alive)  then
       p_out = null_rvec
       return
    end if 
   
    ! Calculate random restitution coefficent
    spd = random_distrib_num(e_dist)
    if (spd < 0) then 
       spd = 0
    else if (spd > 1) then 
       spd = 1
    end if

    ! Calculate random angle
    ang = random_distrib_num(ang_dist)
    if (ang < 0)  then
       ang = 0
    else if (ang > pi) then 
       ang = 1
    end if
    ang = ang*pi
    
    p_out = s*spd*rvec_t(cos(ang),sin(ang))

  end subroutine fixed_ricoch

  ! Ricochet function for bouncing off a flat surface with friction
  ! Initial angular velocity assumed to be zero
  ! and speed_mean
  ! Angular momentum not yet accounted for
  subroutine flat_ricoch(pos, diam, p_in, p_out, alive)
    type(rvec_t), intent(in)  :: pos    ! position
    real(r_kind), intent(in)  :: diam   ! diameter
    type(rvec_t), intent(in)  :: p_in   ! in momentum
    type(rvec_t), intent(out) :: p_out  ! out momentum
    logical, intent(out)      :: alive  ! Did the particle ricochet?

    real(r_kind) :: e
    real(r_kind) :: ang
    real(r_kind) :: t_imp
    real(r_kind), parameter :: I = 0.4 ! Moment of inertia ratio for a sphere
    real(r_kind), parameter :: Ir = I/(1+I) ! used in calculation
    
    type(rvec_t) :: p
    type(rmat_t) :: r ! rotation matrix

    ! Now decide if particle bounces or not 
    alive = random_test(eval_fnct(prob_fnct,length(p_in)/grain_mom_scale))
    if (.not. alive)  then
       p_out = null_rvec
       return
    end if

    p = p_in

    do
       ! Calculate a random angle for the surface
       ang = pi*random_distrib_num(ang_dist) 
       r = rotate_matrix(rang_t(ang))
       
       p = r * p   ! Rotate input frame
       e = random_distrib_num(e_dist) ! Calculate a restitution coefficent
       ! Calculation assumes incoming angular velocity is zero
       p_out%y = -e*p%y 
       t_imp = friction*(1+e)*abs(p%y)            ! Tangential impulse
       if (t_imp>Ir*abs(p%x)) t_imp = Ir*abs(p%x) ! Limit slip case
       p_out%x = p%x - sign(t_imp,p%x)
       
       p_out = p_out * r   ! Rotate back, (transposed equation)
       
       if (p_out%y>=0) then ! Did the particle bounce up ?
          exit  ! Yes we're finished
       else            
          p=p_out  ! No calculate another ricochet
       end if
    end do
  end subroutine flat_ricoch
end module ricochet
