module splash
  use particle
  use track_profile
  use bed
  use grain_var
  use sample
  use maths_const
  use sysdep_io
  use traj_samp
  use maths
  use ricochet
  use random
  use string_param
  use random_distrib

  implicit none
  private

  public :: read_splash_nl    ! Read in parameters controlling splash function
  public :: write_splash_nl   ! Write splash parameters
  public :: init_splash       ! Init splash
  public :: end_splash        ! End splash
  public :: grain_impact      ! Main function. Grain impacts ground

  real(r_kind) :: num_mean     = 0.16
  real(r_kind) :: num_exp      = 1     ! num = num_mean*v^num_exp

  real(r_kind) :: speed_exp    = 0.3   ! vout = v^.3

  type(distrib_pub_t) :: angle = distrib_pub_t("normal",70,10)
  type(distrib_pub_t) :: speed = distrib_pub_t("exponential",0.7,0.7)
  type(distrib_pub_t) :: x_vel = distrib_pub_t("normal",1.1,0.92)
  type(distrib_pub_t) :: y_vel = distrib_pub_t("exponential",1.8,1.8)
  type(distrib_pub_t) :: efficiency = distrib_pub_t("beta",0.2,0.01)


  logical :: limit_mom    = .false. ! Limit momentum
  logical :: poisson      = .true.  ! Use poisson random variable for number
  ! Can a particle that doesn't ricochet splash other particles
  logical :: after_death  = .true.
  real(r_kind) :: ratio = 1.5

  character(len=STRING_LENGTH) :: splash_type = 'cartesian'
  character(len=STRING_LENGTH) :: vel_dep_type = 'diff'

  namelist /splash_nl/splash_type, vel_dep_type, num_mean, num_exp, &
       x_vel, y_vel, angle, speed, speed_exp, efficiency, &
       limit_mom, poisson, after_death, ratio

  integer :: warning_20 = 0
  integer :: warning_limiting = 0
  integer :: warning_mom_ex = 0

  integer, parameter :: UNDECODED_T = -1

  integer, parameter :: NONE_T      = 1
  integer, parameter :: CARTESIAN_T = 2
  integer, parameter :: POLAR_T     = 3
  integer, parameter :: MCEWAN_T    = 4
  integer, parameter :: MCELWAINE_T = 5

  type(param_table_t), parameter :: splash_types(5) = (/&
       param_table_t("none",      NONE_T),&
       param_table_t("cartesian", CARTESIAN_T),&
       param_table_t("polar",     POLAR_T),&
       param_table_t("mcewan",    MCEWAN_T),&
       param_table_t("mcelwaine", MCELWAINE_T)/)


  ! The next lot of constants determine what the number of ejected grains
  ! depends on
  ! 
  integer, parameter :: IN_T   = 1
  integer, parameter :: OUT_T  = 2
  integer, parameter :: DIFF_T = 3
  integer, parameter :: XIN_T   = 4
  integer, parameter :: XOUT_T  = 5
  integer, parameter :: XDIFF_T = 6
  integer, parameter :: YIN_T   = 7
  integer, parameter :: YOUT_T  = 8
  integer, parameter :: YDIFF_T = 9
  integer, parameter :: VDIFF_T = 10
  integer, parameter :: EDIFF_T = 11

  type(param_table_t), parameter :: vel_dep_types(11) = (/&
       param_table_t("in",   IN_T  ),&
       param_table_t("out",  OUT_T ),&
       param_table_t("diff", DIFF_T),&
       param_table_t("xin",  XIN_T  ),&
       param_table_t("xout", XOUT_T ),&
       param_table_t("xdiff",XDIFF_T),&
       param_table_t("yin",  YIN_T  ),&
       param_table_t("yout", YOUT_T ),&
       param_table_t("ydiff",YDIFF_T),&
       param_table_t("vdiff",VDIFF_T),&
       param_table_t("ediff",EDIFF_T)/)

  ! Decoded distribution
  type(distrib_t) :: angle_dist = null_distrib_t
  type(distrib_t) :: speed_dist = null_distrib_t
  type(distrib_t) :: vel_x_dist = null_distrib_t
  type(distrib_t) :: vel_y_dist = null_distrib_t
  type(distrib_t) :: efficiency_dist = null_distrib_t

  integer :: decoded_splash = UNDECODED_T
  integer :: vel_dep = UNDECODED_T          

contains
  subroutine read_splash_nl(fp)
    integer, intent(in) :: fp
    integer iostat
    read(fp, nml=splash_nl,err=1,iostat=iostat)
    return
1   call print_iostat(iostat,std_err,&
         'Error with splash_nl reading file "'//trim(get_name(fp))//'"')
    call error_halt
  end subroutine read_splash_nl
  subroutine write_splash_nl(fp)
    integer, intent(in) :: fp
    write(fp,'(A)')      ' &splash_nl'
    write(fp,'(A)')      '    splash_type  = "' // trim(splash_type) // '"'
    write(fp,'(A)')      '    vel_dep_type = "' // trim(vel_dep_type) // '"'
    write(fp,'(A,G11.4)')'    num_mean     = ', num_mean
    write(fp,'(A,G11.4)')'    num_exp      = ', num_exp
    write(fp,'(A,G11.4)')'    speed_exp    = ', speed_exp
    write(fp,'(A,G11.4)')'    ratio        = ', ratio    
    write(fp,'(A,L1)')   '    after_death  = ', after_death
    write(fp,'(A,L1)')   '    limit_mom    = ', limit_mom
    write(fp,'(A,L1)')   '    poisson      = ', poisson
    write(fp,'(A,G11.4)',advance='no')'    angle         = '
    call write_distrib(fp,angle)
    write(fp,'(A,G11.4)',advance='no')'    speed         = '
    call write_distrib(fp,speed)
    write(fp,'(A,G11.4)',advance='no')'    x_vel         = '
    call write_distrib(fp,x_vel)
    write(fp,'(A,G11.4)',advance='no')'    y_vel         = '
    call write_distrib(fp,y_vel)
    write(fp,'(A,G11.4)',advance='no')'    efficiency        = '
    call write_distrib(fp,efficiency)
    write(fp,'(A)')      ' /'
    write(fp,'(A)')' '
  end subroutine write_splash_nl
  subroutine init_splash()
    type(distrib_pub_t) a
    decoded_splash = decode_string_param(splash_type,splash_types)

    vel_dep = decode_string_param(vel_dep_type,vel_dep_types)

    ! Convert angles to a scale of [0,1]
    a = angle
    a%mean = angle%mean/180
    a%std  = angle%std/180
    angle_dist  = init_distrib(a)
    speed_dist  = init_distrib(speed)
    vel_x_dist  = init_distrib(x_vel)
    vel_y_dist  = init_distrib(y_vel)
    efficiency_dist = init_distrib(efficiency)

   end subroutine init_splash

  subroutine end_splash()
    if (warning_20>0) write(std_err,*)&
         'splash.f90: More than 20 grain ejected', warning_20,'times'
    if (warning_limiting>0) write(std_err,*)&
         'splash.f90: Limiting number of grains', warning_limiting,'times'
    if (warning_mom_ex>0) write(std_err,*)&
         'splash.f90: Momentum exceeded', warning_mom_ex,'times'
  end subroutine end_splash

  ! The main function
  ! Calculate a grain impact with the ground
  ! mgrain - is the impacting grain
  ! ngrain - is a list containing space for new grains
  ! new    - is the number of new grains created
  subroutine grain_impact(mgrain,ngrains,new)
    type(particles), intent(inout) :: mgrain   ! Mother grain
    type(particles), intent(out) :: ngrains(:) ! New grains
    integer, intent(out) :: new                ! Number of new grains

    type(rvec_t) p_in        ! incoming momentum
    type(rvec_t) p_out       ! outgoing momentum

    ! Calculate ricochet angle and velocity
    p_in = mgrain%mass*mgrain%vel
    call grain_ricochet(mgrain%pos, mgrain%diam, p_in, p_out, mgrain%alive)

    if (mgrain%alive) then     ! Mother bounces elastically
       mgrain%vel       = p_out/mgrain%mass
       mgrain%parent    = mgrain%num   ! It's its own parent
       mgrain%origin    = ORIGIN_RICOCHET
       mgrain%start_vel = mgrain%vel
       mgrain%start_pos = mgrain%pos
       mgrain%start_t   = mgrain%time
       mgrain%max_y     = mgrain%pos%y    
       mgrain%acc       = null_rvec
    end if

    if (mgrain%alive .or. after_death) then
       call calc_splash_grains(mgrain, p_in, p_out, ngrains, new)
    else
       new = 0
    end if

    if (mgrain%alive) mgrain%num = new_grain_num() ! Give it a new number
  end subroutine grain_impact

  ! Calculates the number of splashed grains as some funtion of the momentum
  ! Input, output or change in momentum 
  ! all, x or y components of momentum
  ! number is proportional to this as a power law
  subroutine calc_num_standard(p_in,p_out,dp,new)
    type(rvec_t), intent(in)  :: p_in
    type(rvec_t), intent(in)  :: p_out
    real(r_kind), intent(out) :: dp
    integer,      intent(out) :: new

    real(r_kind) myu

    select case (vel_dep)
    case(IN_T)
       dp = length(p_in)
    case(OUT_T)
       dp = length(p_out)
    case(DIFF_T)
       dp = length(p_in)-length(p_out)
    case(XIN_T)
       dp = abs(p_in%x)
    case(XOUT_T)
       dp = abs(p_out%x)
    case(XDIFF_T)
       dp = abs(p_in%x)-abs(p_out%x)
    case(YIN_T)
       dp = abs(p_in%y)
    case(YOUT_T)
       dp = abs(p_out%y)
    case(YDIFF_T)
       dp = abs(p_in%y)-abs(p_out%y)
    case(VDIFF_T)
       dp = length(p_in-p_out)                       ! Change in momentum
    case(EDIFF_T)
       dp = sqrt(length_sqr(p_in)-length_sqr(p_out)) ! Change in energy
    case default      
       call error_halt('splash.f90: Unknown velocity dependence in calc_num_standard')
    end select

    dp = dp/grain_mom_scale

    myu = (num_mean*dp)**num_exp

    if (poisson) then
       new = random_poisson(myu)
    else
       new = floor(myu)
    end if

    ! Sanity checks
    call assert(new<100,'Number of splashed grains is absurd')

  end subroutine calc_num_standard

  ! Calculate the velocity distribution and number of splashed grains
  ! mgrain  - mother grain
  ! dp      - momemtum it has imparted to the bed
  ! ngrains - array for new grains
  ! new     - number of new grains


  subroutine calc_splash_grains(mgrain, p_in, p_out, ngrains, new)
    type(particles), intent(in)    :: mgrain
    type(rvec_t),    intent(in)    :: p_in
    type(rvec_t),    intent(in)    :: p_out
    type(particles), intent(out)   :: ngrains(:)
    integer,         intent(out)   :: new     ! Number of new daughter grains
    real(r_kind) dp

    real(r_kind), allocatable :: mom(:), ang(:)

    type(rvec_t), pointer :: p(:) ! Momenta of new particles

    real(r_kind) rat ! Prevent increase in linear momentum
    integer i,j

    ! Nine possibilities here for velocity dependence

    select case (decoded_splash)
    case (NONE_T)
       new = 0
    case (CARTESIAN_T)
       call calc_num_standard(p_in,p_out,dp,new)
       if (new>0) then
          allocate(p(new))
          p%x = grain_mom_scale*random_distrib_num(vel_x_dist,new)
          p%y = grain_mom_scale*random_distrib_num(vel_y_dist,new)
       end if
    case (POLAR_T)
       call calc_num_standard(p_in,p_out,dp,new)
       if (new>0) then
          allocate(mom(new),ang(new),p(new))
          mom=grain_mom_scale*(dp**speed_exp)*random_distrib_num(speed_dist,new)
          ang = pi*random_distrib_num(angle_dist,new)
          where (ang<0)  ang = 0
          where (ang>pi) ang = pi
          p = mom*rvec(cos(ang), sin(ang))
          deallocate(ang,mom)
       end if
    case (MCELWAINE_T)
       call calc_mcelwaine_splash(p_in,p_out,new,p) 
    case (MCEWAN_T)
       call calc_mcewan_splash(p_in,p_out,new,p) 
    case default
       call error_halt(&
            'splash.f90: Unknown splash function requested')
    end select

    if (new>0) then       
       ! Remove particles that have bounced into the ground
       i = 0
       do j=1,new
          if (p(j)%y>=0) then
             i=i+1
             p(i)=p(j)
          end if
       end do
       new = i
 
       call assert(all(p(1:new)%y>=0),'Splash y vel is negative')
       if (limit_mom) then
          rat = (sum(length(p))+length_sqr(p_out))/length(p_in)
          ! Check momentum of the grains, assuming they all have the same mass
          if (rat >= 1) then
             warning_mom_ex = warning_mom_ex+1
             p = p/rat
             write(*,*)"Momentum exceeded",rat
          end if
       end if
       if (new>size(ngrains)) then
          new = size(ngrains)
          warning_limiting = warning_limiting+1
       end if
       
       call make_daughters(ngrains(1:new), p(1:new), mgrain)
       deallocate(p)
    end if
  end subroutine calc_splash_grains

  subroutine calc_mcelwaine_splash(p_in, p_out, new, p)
    type(rvec_t), intent(in)    :: p_in
    type(rvec_t), intent(in)    :: p_out
    integer,      intent(out)   :: new     ! Number of new daughter grains
    type(rvec_t), pointer       :: p(:)    ! New momenta

    real(r_kind), allocatable :: ang(:)
    type(rvec_t) dp  ! Impulse imparted to the ground
    real(r_kind) de  ! Energy available not scaled by particle mass
    integer i

    de = length_sqr(p_in)-length_sqr(p_out)
    if (de<tiny(de)) then
       new=0
       return
    end if

    de = sqrt(de)/grain_min_mom*random_distrib_num(efficiency_dist)

    if (ratio>1+epsilon(ratio)) then
       new = floor(log(de*(ratio-1))/log(ratio))
    else
       new =  floor(de)
    end if

    if (new<1) then
       new = 0
       return
    end if
    
    call assert(new<100,'Number of splashed grains is absurd')
    ! Calculate direction for splashed particles
    dp = p_in-p_out
    dp%y = -dp%y ! Reflect upwards
    dp = grain_min_mom*dp/length(dp)
    allocate(p(new),ang(new))
    ang = pi*random_distrib_num(angle_dist,new)
    p = rotate(arang_t(ang),dp) * 2*random_uniform(new)*ratio**(/(i,i=0,new-1)/)
    deallocate(ang)
  end subroutine calc_mcelwaine_splash


  ! Create daughter particles using the mother 
  ! momentum has been written to velocity
  subroutine make_daughters(ngrains, p, mgrain)
    type(rvec_t), intent(in)     :: p(:)           ! momentum
    type(particles), intent(out) :: ngrains(size(p))     ! New grains
    type(particles), intent(in)  :: mgrain         ! Mother grain 

    ngrains%vel    = p
    ngrains%origin = ORIGIN_SPLASH
    ngrains%pos    = mgrain%pos
    ngrains%gy     = mgrain%gy
    ngrains%time   = mgrain%time
    ngrains%parent = mgrain%num

    call random_grain_size(ngrains%diam)

    ! Flag mom=.true. explains that momentum not velocity has been specified
    call add_grain(ngrains,mom=.true.) ! Calculate mass, alive flag, number etc
  end subroutine make_daughters

  subroutine calc_mcewan_splash(p_in, p_out, new, p)
    type(rvec_t), intent(in)    :: p_in
    type(rvec_t), intent(in)    :: p_out
    integer,      intent(out)   :: new     ! Number of new daughter grains
    type(rvec_t), pointer       :: p(:)    ! New momenta

    real(r_kind), allocatable :: ang(:),rnd(:),v(:)
    real(r_kind) ldp_max  ! Impulse imparted to the ground
    real(r_kind) ldp_min  ! Impulse imparted to the ground
    real(r_kind) myu,vel

    ldp_max = abs(p_in%x-p_out%x)/grain_mom_scale
    ldp_min = ldp_max-233
    if (ldp_min<0) ldp_min = 0

    myu = 0.075*(ldp_min +(ldp_max-ldp_min) * random_uniform())
    new = nint(myu)
    vel = 0.43*(ldp_min +(ldp_max-ldp_min) * random_uniform())

    allocate(p(1:new),v(1:new),ang(1:new),rnd(1:new))
    rnd = 0.082*random_gamma(real(5.4,r_kind),new)
    v = grain_mom_scale * vel * rnd /sum(rnd)
    ang = pi/180*(125-5*v/grain_mom_scale)*0.082*random_gamma(real(5.4,r_kind),new)
    p = v*rvec(cos(ang),sin(ang))
    deallocate(v,ang,rnd)
  end subroutine calc_mcewan_splash
end module splash
