# 1 "sysdep_io.F90"
! Copyright 1998-1999 Jim McElwaine
! Functions for writing error messages, warning messages and info messages
! As well as asserts and finding unused unit numbers
module sysdep_io
  use precision

  use nag_iostat        ! For NAG compiler we can print error messages
# 10

  implicit none 

!-- Public procedures.
  public :: error_halt  ! Stop the program with an error message  
  public :: info_msg    ! Display an informational message
  public :: warning_msg ! Display a warning message
  public :: error_msg   ! Display an error message
!  public :: assign_lun  ! Assign a file unit number
!  public :: free_lun    ! Unassign a file unit number
  public :: assert      ! error message and halt if assertion is not true
  public :: warn_assert ! warning message if assertion is not true
  public :: err_open    ! open command but with error handling 
  public :: err_close   ! open command but with error handling
  public :: get_name    ! return the file name (if any) of a file unit

  interface assert
     module procedure l_assert, r_assert, equ_assert, &
          ar_assert, aequ_assert, al_assert
  end interface

# 34

  logical,parameter,public :: DEBUG_FLAG = .false.
  logical,parameter,public :: WARN_FLAG = .false.


  integer, parameter, public :: FILE_NAME_LEN = 50
  integer, parameter, public :: STR_LEN = 500

!-- standard files
  integer, public :: std_in  = 5
  integer, public :: std_out = 6
  integer, public :: std_err = 6
  integer, public :: std_log      ! Logfile for messages etc

! Handling of file unit numbers handling
  integer, parameter :: min_lun = 10
  integer, parameter :: max_lun = 99
  logical, private   :: used_lun(min_lun:max_lun) = .false.

contains
  subroutine error_halt (message)
    character(len=*), intent(in), optional :: message 
    if (present(message)) call error_msg(message)
    call error_stop
  end subroutine error_halt

  subroutine error_msg (message)
    character(len=*), intent(in) :: message
    write (std_err, *) 'Error: ', trim(message)
  end subroutine error_msg

  subroutine warning_msg(message)
    character(len=*), intent(in) :: message
    write (std_err, *) 'Warning: ', trim(message)
  end subroutine warning_msg

  subroutine l_assert(condition, message)
    character(len=*), intent(in) :: message
    logical, intent(in)       :: condition
    if (.not.DEBUG_FLAG) return
    if (.not.condition) call error_halt(message)
  end subroutine l_assert

  subroutine r_assert(condition, message)
    character(len=*), intent(in) :: message
    real(r_kind), intent(in)     :: condition
    character(len=str_len) :: es
    if (.not.DEBUG_FLAG) return
    if (abs(condition) > 10*epsilon(1.)) then
       write(es,*) condition, 'over epsilon', condition/epsilon(1.)
       call error_halt(trim(message) // trim(es))
       elseif (abs(condition) > epsilon(1.)) then
       write(es,*) condition, 'over epsilon', condition/epsilon(1.)
       call warning_msg(trim(message) // trim(es))
    end if
  end subroutine r_assert 

  subroutine equ_assert(a,b, message)
    character(len=*), intent(in) :: message
    real(r_kind), intent(in)     :: a,b
    real(r_kind) c
    character(len=str_len) :: es
    if (.not.DEBUG_FLAG) return
    c = abs(a-b)
    if(c==0) then 
       return
       elseif (c < epsilon(1.)) then
       write(es,*) a,'does not equal',b, 'over epsilon', c/epsilon(1.)
       call warning_msg(trim(message) // trim(es))
       elseif (c < 100*epsilon(1.)) then
       write(es,*) a,'does not equal',b, 'over epsilon', c/epsilon(1.)
       call warning_msg(trim(message) // trim(es))
    else
       write(es,*) a,'does not equal',b, 'over epsilon', c/epsilon(1.)
       call error_halt(trim(message) // trim(es))
    end if
  end subroutine equ_assert 

  subroutine al_assert(condition, message)
    character(len=*), intent(in) :: message
    logical, intent(in)       :: condition(:)
    if (.not.DEBUG_FLAG) return
    if (any(.not.condition)) call error_halt(message)
  end subroutine al_assert

  subroutine ar_assert(condition, message)
    character(len=*), intent(in) :: message
    real(r_kind), intent(in)     :: condition(:)
    call assert(maxval(abs(condition)),message)
  end subroutine ar_assert

  subroutine aequ_assert(a,b, message)
    character(len=*), intent(in) :: message
    real(r_kind), intent(in)     :: a(:),b(:)
    real(r_kind) c
    character(len=str_len) :: es
    if (.not.DEBUG_FLAG) return
    c = maxval(abs(a-b))
    if(c==0) then 
       return
       elseif (c < epsilon(1.)) then
       write(es,*) a,'does not equal',b, 'over epsilon', c/epsilon(1.)
       call warning_msg(trim(message) // trim(es))
       elseif (c < 100*epsilon(1.)) then
       write(es,*) a,'does not equal',b, 'over epsilon', c/epsilon(1.)
       call warning_msg(trim(message) // trim(es))
    else
       write(es,*) a,'does not equal',b, 'over epsilon', c/epsilon(1.)
       call error_halt(trim(message) // trim(es))
    end if
  end subroutine aequ_assert

  subroutine warn_assert(condition, message)
    character(len=*), intent(in) :: message
    logical, intent(in)       :: condition
    if (.not.WARN_FLAG) return
    if (.not.condition) call warning_msg(message)
  end subroutine warn_assert

  subroutine info_msg(message,flag,data)
    character(len=*), intent(in) :: message
    logical, optional :: flag
    real(r_kind), optional :: data
    if (present(flag)) then 
       if (.not.flag) return
       write (std_out, *) 'Debugging info: ', trim(message)
    else
       write (std_out, *) 'Info: ', trim(message)
    end if
    if (present(data)) write (std_out, *) 'Data: ', data
  end subroutine info_msg

  subroutine assign_lun (lun)     ! Return a free unit number
    integer, intent(out) :: lun 
    integer :: iostat
    logical :: used
    do lun = min_lun , max_lun
       if (used_lun(lun)) cycle
       inquire(unit=lun, opened=used, iostat=iostat)
       if (iostat == 0 .and. .not.used) then
          used_lun(lun) = .true.
          return
       end if
    end do
    call error_halt('No luns available in assign_lun')
  end subroutine assign_lun

  subroutine free_lun(lun)
    integer, intent(in) :: lun
    if (used_lun(lun)) then
       used_lun(lun) = .false.
    else
       call error_halt('Freeing an unused lun in free_lun')
    end if
  end subroutine free_lun

  subroutine error_stop
# 197

    stop 'Program error detected. Terminating execution'

  end subroutine error_stop 
! Open a file with error messages

  subroutine err_open(unit,file,status,access,form,&
       recl,blank,position,action,delim,pad)
    integer,                    intent(out):: unit
    character(len=*), optional, intent(in) :: file  
    character(len=*), optional, intent(in) :: status
    character(len=*), optional, intent(in) :: access
    character(len=*), optional, intent(in) :: form
    integer,          optional, intent(in) :: recl
    character(len=*), optional, intent(in) :: blank
    character(len=*), optional, intent(in) :: position
    character(len=*), optional, intent(in) :: action
    character(len=*), optional, intent(in) :: delim
    character(len=*), optional, intent(in) :: pad
    integer  :: iostat
    call assign_lun(unit)
    open(unit=unit,iostat=iostat,err=1,file=file,status=status,&
         access=access,form=form,recl=recl,blank=blank,&
         position=position,action=action,delim=delim,pad=pad)
    return
1   call print_iostat(iostat,std_err,'Error opening file "'//file//'"')
    call error_halt
  end subroutine err_open

  subroutine err_close(unit,status)
    integer,                    intent(in) :: unit
    character(len=*), optional, intent(in) :: status
    integer  :: iostat
    close(unit,iostat=iostat,err=1,status=status)
    call free_lun(unit)
    return
1   call print_iostat(iostat,std_err,&
         'Error closing file "'//trim(get_name(unit))//'"')
    call error_halt
  end subroutine err_close

  function get_name(fp) result(name)
    integer, intent(in) :: fp
    character(len=FILE_NAME_LEN) name
    logical named
    integer iostat
    inquire(fp,name=name,named=named,err=1,iostat=iostat)
    if (.not.named) then
       if (fp==std_out) then
          name='std_out'
       else if (fp==std_err) then
          name='std_err'
       else if (fp==std_in) then
          name='std_in'
       else
          name='no name'
       end if
    end if
    return
1   call print_iostat(iostat,std_err,'Error in sysdep_io.f90 get_name()')
    call error_halt
  end function get_name
end module sysdep_io
