module track_grains
  use bed
  use discrete
  use entrainment
  use grain_var
  use impact_stats
  use particle
  use precision
  use ricochet
  use sample
  use splash
  use stats
  use sysdep_io
  use track_profile
  use traj_samp
  use vector
  use wind
  use rksuite_90
  use locate
  use drag
  use maths
  use profile
  use random
  use random_distrib

  implicit none
  private

  public :: track_fnct         ! Integrate a particle through one big time step
  public :: read_track_nl
  public :: write_track_nl
  public :: init_track_grains
  public :: lacalc_fnct! Calculate all particle movement for a large time step
  public :: disp_part_stats    ! Display information about the grains
  public :: read_track_State   ! Read the grain information from a file
  public :: write_track_State  ! Write the grain information to a file
  public :: display_track_step ! display information evry step
  public :: write_track_step   ! write out info to data file
  public :: get_track_step     ! return information every step
  public :: end_track

  logical, parameter :: use_rksuite = .true.

  integer, allocatable :: test_tracks(:) ! Which tracks we are tracking
  integer :: test_tracks_fp ! File pointer for track file

  type(rvec_t) :: e_grav ! Effective bravity adjusted for buoyancy

  ! Variables for track_nl
  integer      :: ntest        = 0               ! How many tracks to test
  type(rvec_t) :: gravity      = rvec_t(0,-9.81) ! Gravity 
  integer      :: max_grains   = 1000            ! maximum number of grains
  real(r_kind) :: tolerance    = 1e-4            ! Tolerance for o.d.e. solver
  real(r_kind) :: threshold(4) = (/1,1,1,1/)     ! Thresholds for o.d.e. solver 
  character    :: ode_method   = 'l'             ! 'l'=low, 'm'=medium, 'h'=high
  character(len=FILE_NAME_LEN) :: grains_fn = "grains_state.bin" ! File name for grains
  logical      :: load_state = .false.          ! Load grain state from a file  
  logical      :: dynamic    = .true.  ! Time dependent or time independent
  logical      :: seed_grain = .false. ! add non aero grains

  integer      :: initial_num = 0 ! How many grains to start off with
  ! Velocity distribution for initial grains
  type(distrib_pub_t) :: x_vel = distrib_pub_t("normal",1.1,0.92) 
  type(distrib_pub_t) :: y_vel = distrib_pub_t("exponential",1.8,1.8)


  namelist /track_nl/ ntest, gravity, max_grains, dynamic, &
       tolerance, threshold, ode_method, grains_fn, load_state, seed_grain, &
       initial_num, x_vel, y_vel
 
  type(particles), allocatable:: grains(:)    ! Master list of particles
  integer            :: num_p                 ! Number of particles
  integer            :: old_num_p             ! Old number of particles
  logical, parameter :: test_lacalc = .false. ! Extra debugging info

  type(particles), private :: cur_g ! Current grain we're tracking

  type(distrib_t) :: vel_x_dist = null_distrib_t
  type(distrib_t) :: vel_y_dist = null_distrib_t

contains
  subroutine end_track(dirn)
    character(len=*), intent(in) :: dirn
    call write_track_state(dirn)
    deallocate(grains,test_tracks)
    initial_num = 0                  ! We don't need initial grains again
    load_state = .true.              ! Prepare name list for load_state
    call end_impact_stats() ! Display total impact statistics
  end subroutine end_track
  subroutine read_track_state(dirn)
    character(len=*), intent(in) :: dirn
    integer fp
    integer iostat
    call assign_lun(fp)
    call err_open(fp,file=trim(dirn)//trim(grains_fn),form='unformatted',&
         status='old',action='read')
    read(fp,err=1,iostat=iostat)num_p
    if (num_p>size(grains)) then
       write(std_err,*)'file '//trim(dirn)//trim(grains_fn)//&
            ' has too many grains'
       num_p = size(grains)
    end if
    write(std_out,*)'Reading',num_p,'grains from "',&
         trim(dirn)//trim(grains_fn),"'"
    read(fp,err=1,iostat=iostat)grains(1:num_p)
    call read_trace_state(fp) !read all the partial profile information
    call read_impact_state(fp)
    call err_close(fp);
    return
1   call print_iostat(iostat,std_err,'Error reading file "'//trim(get_name(fp))//'"')
    call error_halt
  end subroutine read_track_state
  subroutine write_track_state(dirn)
    character(len=*), intent(in) :: dirn
    integer fp
    call assign_lun(fp)
    call err_open(fp,file=trim(dirn)//trim(grains_fn),form='unformatted',&
         status='replace', action='write')
    write(fp)num_p
    write(fp)grains(1:num_p)
    call write_trace_state(fp)! Save all the partial profile information
    call write_impact_state(fp) ! Save all impact data
    call err_close(fp);
  end subroutine write_track_state
  subroutine disp_part_stats(fp)
    integer, intent(in) :: fp
    write(fp,*)'Num p :',num_p,', change',num_p-old_num_p
  end subroutine disp_part_stats
  subroutine disp_grain_info(fp,g)
    integer, intent(in) :: fp
    type(particles), intent(in) :: g(:)
    type(stats_r_t) :: ry
    type(stats_r_t) :: vy
    call stats_clear(ry)
    call stats_clear(vy)
    call stats_add(ry,g%pos%y)
    call stats_add(vy,g%vel%y)
    call stats_print(fp,ry,'Grain heights')
    call stats_print(fp,vy,'Grain vertical velocities')
  end subroutine disp_grain_info
  subroutine dump_track(cur_g)
    type(particles), intent(in) :: cur_g
    write(test_tracks_fp,*)cur_g%num,cur_g%time,cur_g%pos,cur_g%vel,cur_g%acc
  end subroutine dump_track
  subroutine init_track_grains(dirn)
    character(len=*), intent(in) :: dirn
    integer i    
    call init_track_profile()      ! initialize vertical profiles
    call init_impact_stats()       ! initialize impact statistics
    allocate(grains(max_grains),test_tracks(ntest))
    grains%alive = .false.
    test_tracks = (/(i,i=1,ntest)/)   ! Just test first grains Possibly length zero
    if (ntest>0) then
       write(std_out,*)'Writing track data to ' // trim(dirn) // 'track.dat'
       call assign_lun(test_tracks_fp)
       call err_open(test_tracks_fp,file=trim(dirn) // 'track.dat',status='replace')
    end if
    ! Adjust gravity for buoyancy
    e_grav = (1-air_density/grain_dens)*gravity
    vel_x_dist = init_distrib(x_vel)
    vel_y_dist = init_distrib(y_vel)

    if (load_state) then !Load state from a file
       call read_track_state(dirn)
    else
       num_p=0
    end if    
    ! Now create extra initial grains if requested
    if (num_p+initial_num>max_grains) initial_num = max_grains - num_p
    call make_initial_grains(grains(num_p+1:num_p+initial_num))
    num_p=num_p+initial_num
 
   end subroutine init_track_grains
   subroutine read_track_nl(fp)
     integer, intent(in) :: fp
     integer iostat
     read(fp, nml=track_nl,err=1,iostat=iostat)
     return
1    call print_iostat(iostat,std_err,&
          'Error with track_nl reading file "'//trim(get_name(fp))//'"')
     call error_halt
   end subroutine read_track_nl
  subroutine write_track_nl(fp)
    integer, intent(in) :: fp
    write(fp,'(A)')       ' &track_nl'
    write(fp,'(A,I5)')    '    ntest       = ', ntest
    write(fp,'(A,2G12.4)')'    gravity     =',  gravity
    write(fp,'(A,I6)')    '    max_grains  = ', max_grains
    write(fp,'(A,G11.4)') '    tolerance   = ', tolerance
    write(fp,'(A,4G12.4)')'    threshold   =',  threshold
    write(fp,'(A,A1)')    '    ode_method  = "' // ode_method // '"'
    write(fp,'(A,A)')     '    grains_fn   = "' // trim(grains_fn) // '"'
    write(fp,'(A,L1)')    '    load_state  = ', load_state
    write(fp,'(A,L1)')    '    dynamic     = ', dynamic
    write(fp,'(A,L1)')    '    seed_grain  = ', seed_grain
    write(fp,'(A,I6)')    '    initial_num = ', initial_num
    write(fp,'(A,G11.4)',advance='no')'    x_vel       = '
    call write_distrib(fp,x_vel)
    write(fp,'(A,G11.4)',advance='no')'    y_vel       = '
    call write_distrib(fp,y_vel)
    write(fp,'(A)')       ' /'
    write(fp,'(A)')' '
  end subroutine write_track_nl

  ! Integrate all grains forward in time possibly increasing grain numbers
  ! Grains(1:num_p) contains live particles 
  ! num_p can be changed in the routine but the list must stay
  ! continuous
  ! start_time - program time at start of step
  ! tstep      - how far forward in time to integrate
  ! grain      - list of grains
  ! num_p      - current number of grains
  subroutine track_fnct(start_time, tstep, grains, num_p)
    real(r_kind), intent(in)       :: start_time
    real(r_kind), intent(in)       :: tstep
    type(particles), intent(inout) :: grains(:)
    integer, intent(inout)         :: num_p

    type(rvec_t) oldpos, oldvel
    integer oldj,ig
    integer  ejected
    logical impfl ! Impact flag
    logical tracking ! Are we tracking this grain
    real(r_kind) rel_dens
    real(r_kind) end_time
    real(r_kind) oldtime
    real(r_kind) tt  ! Time offset to actual collision time

    real(r_kind) f(4)
    real(r_kind) x(4)
    integer flag
    type(rk_comm_real_1d) :: rk_comm ! Communicator for o.d.e solver

    end_time = start_time + tstep

    rel_dens=air_density/grain_dens

    ig = 1
    ! Within the loop the grain is accessed via the variable g 
    ! not the list element grains(ig)

    call setup(rk_comm,cur_g%time,x,end_time, &
         tolerance,threshold,method=ode_method,task='S')

    all_grains: do            ! track all grains in the list
    
    ! Find a grain to integrate
      do 
         if (ig>num_p) exit all_grains ! num_p can change in the loop
         if (grains(ig)%time<end_time) exit
         ig = ig + 1  ! Look at next grain 
      end do

      cur_g = grains(ig) ! grain we're working with
      
       
       if (use_rksuite) then
          ! Prepare for the integration
          
          x = (/cur_g%pos%x, cur_g%pos%y, cur_g%vel%x, cur_g%vel%y/)
!   call setup(rk_comm,cur_g%time,x,end_time, &
!         tolerance,threshold,method=ode_method,task='S')
          call reset_rk(rk_comm,cur_g%time,end_time,x)
       else
       end if

       ! Are we recording all tracking information for this grains
       if (any(test_tracks==ig)) then
          tracking = .true.
       else
          tracking = .false.
       end if

       impfl=.false.
       cur_g%gy = lookup(cur_g%pos%y,grid_y,cur_g%gy) 
       if (tracking) call dump_track(cur_g)
       
       do
          
          if (cur_g%time >= end_time) exit
 
          if (use_rksuite) then
             oldtime = cur_g%time
             oldvel  = cur_g%vel
             oldpos  = cur_g%pos
             oldj    = cur_g%gy
             
             ! We must fix step_integrate to stop when y=0
             x = (/cur_g%pos%x, cur_g%pos%y, cur_g%vel%x, cur_g%vel%y/)
             f = (/x(3:4),cur_g%acc%x,cur_g%acc%y/)
             call step_integrate(rk_comm,part_force_fnct,cur_g%time,x,f,flag)
             cur_g%pos  = rvec_t(x(1),x(2))
             cur_g%vel  = rvec_t(x(3),x(4))
             cur_g%acc  = rvec_t(f(3),f(4))
          else
          end if


          ! Translate back in time to impact position
          if (oldvel%y >= 0 .and. cur_g%vel%y<0) then
             cur_g%max_y = highest_point(cur_g%time-oldtime, &
                  oldpos%y,cur_g%pos%y,oldvel%y,cur_g%vel%y)
          end if

          ! Impact,  below the bottom of the grid
          ! Note that acceleration is zero below bottom of grid
          if (cur_g%pos%y <= 0) then
             tt = cur_g%pos%y/cur_g%vel%y
             cur_g%time  = cur_g%time-tt
             cur_g%pos%x = cur_g%pos%x - cur_g%vel%x*tt
             cur_g%pos%y = 0 ! same as cur_g%pos%y - cur_g%vel%y*tt 
             impfl=.true.  
          end if

          ! Keep track of highest point
          ! Find highest point of cubic fit

          if (tracking) call dump_track(cur_g)

          
          cur_g%gy  = lookup(cur_g%pos%y,grid_y,oldj) 
          call assert(cur_g%gy>=1.and.cur_g%gy<size(grid_y),'Particle out of grid')        
            
          ! Has the particle hit the deck ?
          if (impfl) then
             ! This trajectory has now ended so save it
             call delete_grain(cur_g) ! Save grain trajectory info
             ! If it Ricochets however maybe it can still live ...  
             call grain_impact(cur_g,grains(num_p+1:),ejected)
             call add_impact_stats(ejected,cur_g%alive)
             num_p = num_p+ejected 
             call assert(num_p<=size(grains),'Splash has produced too many grains')
          else
             call add_track_profile(oldj,cur_g%gy,oldpos%y,cur_g%pos%y,&
                  cur_g%pos%x-oldpos%x, cur_g%vel%x-oldvel%x, cur_g%mass, &
                  cur_g%time-oldtime)
          end if
          
          ! Keep the list continous if the particle has changed
          ! Be careful since g is aliased to g(ig) 
          if(.not. cur_g%alive) then
             if (ig/=num_p) then
                cur_g=grains(num_p)
                grains(num_p)%alive=.false.
             end if
             num_p = num_p-1
          end if
          if (impfl) exit  ! A grain impacted so restart iteration
       end do  ! Time stepping loop
       grains(ig) = cur_g  ! Put it back in the list 
    end do all_grains ! Loop through particles
    call collect_garbage(rk_comm)
   end subroutine track_fnct

  ! Note that this function requires cur_g to be set  
  ! This contains things like grain mass and drag coefficents
  function part_force_fnct(t,x) result(f)
    real(r_kind), intent(in)      :: t     ! Time 
    real(r_kind), intent(in)      :: x(:)  ! position and velocity
    real(r_kind)                  :: f(size(x))  ! accleration and velocity

    type(rvec_t) vel
    type(rvec_t) pos
    type(rvec_t) acc
    type(rvec_t) rel ! velocity relative to air
     

    if (x(2)<0) then ! Don't waste our time integrating here
       f(1:2) = x(3:4) ! Keep a constant velocity
       f(3:4) = 0      !(/cur_g%acc%x, cur_g%acc%y/)
       return
    end if

    pos = rvec_t(x(1),x(2))  ! Position
    vel = rvec_t(x(3),x(4))  ! Velocity

    ! Calculate relative velocity
    rel%x = air_vel(pos%y,t)-vel%x
    rel%y = -vel%y 

    acc = e_grav + rel * (cur_g%ff * drag_coeff(cur_g%rc*length(rel)))

    f   = (/vel%x, vel%y,  acc%x, acc%y/)

    !          urel%x=uint%x-uold%x
    !          urel%y=uint%y-uold%y
    !----           added mass plus pressure term (accelerations)
    !         add_pres%x = rel_dens*(1.5*(uint%x-uoint%x)/tstep-0.5*cur_g%acc%x)
    !         add_pres%y = rel_dens*(-0.5*cur_g%acc%y)
  end function part_force_fnct
    
  ! Executes one large time step for all the grains
  ! Adds aerodynamicaly entrained grains
  ! Clears all trace and sample variables forone t
  ! Executue
  subroutine lacalc_fnct(time,tstep)
    real(r_kind), intent(in)  :: time           ! Program time
    real(r_kind), intent(in)  :: tstep          ! How afar to go forward
    
    integer aero 
    !Clear fluxes, number densities, etc
    call clear_sample();     ! Clear sample counts
    call aero_entran(get_b_stress(),time, tstep, aero, grains(num_p+1:))
    num_p = num_p+aero
    if (num_p == 0 .and. seed_grain) then ! drop a grain from the top
       call drop_grain(time, tstep, grains(1:1))
       num_p=1
    end if
    call add_aero_total_stats(aero,num_p)!Set the aero num and total num of grains
    
    ! Integrate all the particle forward in time
    call track_fnct(time,tstep,grains,num_p) 

    call finish_track_profile(tstep) ! End the profiling for the track module
    call finish_impact_stats(tstep)  ! Add on the collection of impact stats
  end subroutine lacalc_fnct
  subroutine comb_grains(g,n) ! Currently unused
    type(particles), intent(inout) :: g(:)
    integer, intent(out) :: n
    integer i,j
    !
    !---  'comb' the list (fill the wholes)
    !diam == 0 means there is no grain
    i = 1
    j = size(g)
    do
       if (.not.g(i)%alive .and. g(j)%alive) then
          g(i)=g(j)
          g(j)%diam=0.        ! free the `j' place
          i=i+1
          j=j-1
       endif
       if (g(i)%alive) i=i+1
       if (.not.g(j)%alive) j=j-1
       if (i>j) exit
    end do
    n = j
  end subroutine comb_grains  
  function get_track_step()
    real(r_kind) get_track_step(6)
    get_track_step(1:4) = get_grain_profile()
    get_track_step(5:6) = get_v_fluxes()
  end function get_track_step
  subroutine display_track_step(fp)
    integer, intent(in) :: fp
    call disp_grain_profile(fp)
    call disp_v_fluxes(fp)
    write(fp,*)
  end subroutine display_track_step  ! Set class for each grain
  subroutine write_track_step(fp)
    integer, intent(in) :: fp
    call write_grain_profile(fp)
    call write_v_fluxes(fp)
  end subroutine write_track_step  ! Set class for each grain
  subroutine drop_grain(t,dt,g)
    real(r_kind), intent(in) :: t
    real(r_kind), intent(in) :: dt
    type(particles), intent(out) :: g(:)     ! New grains

    g%vel    = null_rvec
    g%origin = ORIGIN_DROP
    g%pos    = rvec_t(0,grid_height)
    g%gy     = lookup(cur_g%pos%y,grid_y)
    g%time   = t+dt*random_uniform(size(g))!Start them uniformly on the interval
    g%parent = 0

    call random_grain_size(g%diam)
    call add_grain(g)              ! Now calculate mass, drag coefficents etc
  end subroutine drop_grain
  ! Make grains with a distribution of velocities on the surface
  subroutine make_initial_grains(g)
    type(particles), intent(out) :: g(:)     ! New grains
    integer n
    n = size(g)
    g%vel%x = grain_vel_scale*random_distrib_num(vel_x_dist,n)
    g%vel%y = grain_vel_scale*random_distrib_num(vel_y_dist,n)
    g%origin = ORIGIN_INITIAL   ! 
    call random_bed_pos(g%pos)     ! Give them a random initial position
    g%gy=1
    g%time   = 0             !Start them uniformly on the interval
    g%parent = 0
    call random_grain_size(g%diam)
    call add_grain(g)              ! Now calculate mass, drag coefficents etc
  end subroutine make_initial_grains
end module track_grains
