module track_profile
  use discrete
  use precision
  
  use sysdep_io
  use maths
  implicit none
  private

  public :: init_track_profile  ! Called at start of program
  public :: end_track_profile   ! Called at end of program
  public :: start_track_profile ! Called at start of time step
  public :: finish_track_profile ! Called at end of time step
  public :: write_track_profile ! Write a profile
  public :: add_rugby_x         ! Add statistic to the rugby count
  public :: display_rugby_x     
  public :: add_track_profile      ! Display the total momnetum flux
  public :: get_part_force      ! Particle force for current time step
  public :: get_grain_profile
  public :: disp_grain_profile
  public :: write_grain_profile
  public :: write_trace_state
  public :: read_trace_state

  ! Values for one time step
  real(r_kind), allocatable :: num_prof(:)     ! Number density
  real(r_kind), allocatable :: mass_prof(:)    ! Mass density
  real(r_kind), allocatable :: mom_prof(:)     ! Momentum density
  real(r_kind), allocatable :: imp_prof(:)     ! Impulse profile = grain stress
  real(r_kind), allocatable :: last_imp_prof(:)! Two moving point average

  ! Values averaged over several time steps
  real(r_kind), allocatable :: t_num_prof(:)     ! Number density
  real(r_kind), allocatable :: t_mass_prof(:)    ! Mass density
  real(r_kind), allocatable :: t_mom_prof(:)     ! Momentum density
  real(r_kind), allocatable :: t_imp_prof(:)     ! Impulse profile = grain stress

  real(r_kind), allocatable :: rugby_mass(:)   ! Count mass moving past a point
  real(r_kind), allocatable :: rugby_num(:)    ! Count number moving past a point    
  real(r_kind) :: elap_time
  real(r_kind) :: step_time
  real(r_kind) :: last_step_time

contains
  subroutine end_track_profile()
    deallocate(num_prof,mass_prof,mom_prof,imp_prof,&
         t_num_prof,t_mass_prof,t_mom_prof,t_imp_prof,&
         last_imp_prof,rugby_num,rugby_mass)
  end subroutine end_track_profile
  subroutine init_track_profile()
    allocate(num_prof(ngr),mass_prof(ngr),mom_prof(ngr),imp_prof(ngr),&
         t_num_prof(ngr),t_mass_prof(ngr),t_mom_prof(ngr),t_imp_prof(ngr),&
         last_imp_prof(ngr),rugby_num(ngr),rugby_mass(ngr))
    call start_track_profile(.true.)
    step_time=0
    elap_time=0
    call start_track_profile(.true.) ! Zero everything
  end subroutine init_track_profile
  ! Called at the beginning of a step
  subroutine start_track_profile(reset)
    logical, intent(in) :: reset      ! Whether or not to reset counters
    if (reset) then
       elap_time = 0
       t_num_prof  = 0
       t_mass_prof = 0
       t_mom_prof  = 0
       t_imp_prof  = 0
    end if
    last_imp_prof = imp_prof
    last_step_time = step_time

    num_prof  = 0
    mass_prof = 0
    mom_prof  = 0
    imp_prof  = 0
  end subroutine start_track_profile
  subroutine finish_track_profile(dt)
    real(r_kind) :: dt ! How much time elapsed this step
    ! Add on totals
    t_num_prof  = t_num_prof  + num_prof
    t_mass_prof = t_mass_prof + mass_prof
    t_mom_prof  = t_mom_prof  + mom_prof 
    t_imp_prof  = t_imp_prof  + imp_prof 

    elap_time=elap_time+dt
    step_time = dt
  end subroutine finish_track_profile
  ! Called at the beginning of a step

  ! The trace has finished Normalize everything
  subroutine write_trace_state(fp)
    integer, intent(in) :: fp
    write(fp)elap_time,step_time,last_step_time,t_num_prof,&
         t_mass_prof, t_mom_prof, t_imp_prof,last_imp_prof
  end subroutine write_trace_state
  subroutine read_trace_state(fp)
    integer, intent(in) :: fp
    integer iostat
    read(fp,err=1,iostat=iostat)elap_time,step_time,last_step_time,&
         t_num_prof, t_mass_prof, t_mom_prof, t_imp_prof,last_imp_prof
    return
1   call print_iostat(iostat,std_err,'Error reading file "'//trim(get_name(fp))//'"')
    call error_halt
  end subroutine read_trace_state
  subroutine write_track_profile(fp)
    integer, intent(in) :: fp
    if (elap_time==0) then
       write(fp)0*num_prof,0*mass_prof,0*mom_prof,0*imp_prof
    else
       write(fp)t_num_prof/(elap_time*grid_vol),&
            t_mass_prof/(elap_time*grid_vol),&
            t_mom_prof/(elap_time*grid_vol),&
            t_imp_prof/(elap_time*grid_vol)
    end if
  end subroutine write_track_profile
  function get_grain_profile() result(r)
    real(r_kind) :: r(4) 
    if (step_time==0) then  
       r = 0
    else 
       r = (/sum(num_prof),sum(mass_prof),sum(mom_prof),sum(imp_prof)/) &
            / (step_time*grid_area)
    end if
  end function get_grain_profile
  subroutine disp_grain_profile(fp)
    integer, intent(in) :: fp
    if (step_time==0) then
       write(fp,'(1X,A)',advance='no')'step time is zero'
    else
       write(fp,'(A,I5,3(A,F7.4))',advance='no')&
            ' num',   nint(sum(num_prof) /  step_time),&
            ' mass',  sum(mass_prof)/ (step_time*grid_area),&
            ' flux',  sum(mom_prof) / (step_time*grid_area),&
            ' stress',sum(imp_prof) / (step_time*grid_area)
    end if
  end subroutine disp_grain_profile
  subroutine write_grain_profile(fp)
    integer, intent(in) :: fp
    if (step_time==0) then
       write(fp,'(4(ES11.4))',advance='no')0.,0.,0.,0.
    else
       write(fp,'(4(ES11.4))',advance='no')&
            (/sum(num_prof),sum(mass_prof), sum(mom_prof),sum(imp_prof)/)&
            /(step_time*grid_area)
    end if
  end subroutine write_grain_profile

  ! record parameters for the mid point of the track
  ! Count the particles as they move of the bed and wrap around
  subroutine add_rugby_x(j,grain_mass)
    integer, intent(in) :: j
    real(r_kind), intent(in) :: grain_mass
    rugby_mass(j) = rugby_mass(j)+grain_mass  ! Mass, rugby count
    rugby_num(j)  = rugby_num(j)+1            ! Rugby number
  end subroutine add_rugby_x
  subroutine display_rugby_x(fp)
    integer, intent(in) :: fp
    write(fp,*)'Rugby count :',sum(rugby_num)
  end subroutine display_rugby_x
  subroutine add_track_profile(oldj,newj,oldy,newy,dx,dv,mass,dt)
    integer, intent(in) :: oldj
    integer, intent(in) :: newj
    real(r_kind), intent(in) :: oldy
    real(r_kind), intent(in) :: newy
    real(r_kind), intent(in) :: dx
    real(r_kind), intent(in) :: dv
    real(r_kind), intent(in) :: mass
    real(r_kind), intent(in) :: dt

    real(r_kind) imp
    real(r_kind) mom
    real(r_kind) maxy
    real(r_kind) miny
    real(r_kind), allocatable :: fac(:)
    integer il,iu

    imp = dv*mass ! Impulse
    mom = dx*mass ! Momentum - max flux

    if (newj==oldj) then
       imp_prof(oldj)  = imp_prof(oldj)  + imp
       mom_prof(oldj)  = mom_prof(oldj)  + mom
       mass_prof(oldj) = mass_prof(oldj) + mass*dt
       num_prof(oldj)  = num_prof(oldj)  + dt
    else
       if (newj>oldj) then 
          iu = newj
          il = oldj
          maxy = newy
          miny = oldy
       else 
          iu = oldj
          il = newj
          maxy = oldy
          miny = newy
       end if
       allocate(fac(iu-il+1))
       fac(1)           = (grid_y(il+1)-miny)/(maxy-miny)
       fac(2:iu-il)     =  grid_dy(il+1:iu-1)/(maxy-miny)
       fac(iu-il+1)     = (maxy-grid_y(iu))  /(maxy-miny)

       imp_prof(il:iu)  = imp_prof(il:iu)  + fac*imp
       mom_prof(il:iu)  = mom_prof(il:iu)  + fac*mom
       mass_prof(il:iu) = mass_prof(il:iu) + fac*mass*dt
       num_prof(il:iu)  = num_prof(il:iu)  + fac*dt

       call assert(sum(fac)-1_r_kind,'Scaling does not sum to one')
       call assert(all(fac>=0) .and. all(fac<=1),'Scaling wrong')
       deallocate(fac)
    end if
  end subroutine add_track_profile
  function get_part_force(average) result(f)
    real(r_kind) :: f(ngr)
    logical, intent(in) :: average
    if (average) then
       if (step_time+last_step_time == 0) then
          f = 0
       else
          f=(imp_prof+last_imp_prof)/(step_time+last_step_time)
       end if
    else
       if (step_time == 0) then
          f = 0
       else
          f = imp_prof/step_time
       end if
    end if
  end function get_part_force
end module track_profile
