! Module for sampling trajectories
! Includes creation and destruction function for grains

module traj_samp
  use sysdep_io
  use precision
  use particle
  use wind
  private

  public :: delete_grain  ! Called when a grain dies, save trajectory info
  public :: add_grain     ! Calculate auxiallry constants for a grain
  public :: new_grain_num ! Returns a number for a new grain
  public :: read_traj_samp_nl ! Read parameters controlling trajectories
  public :: write_traj_samp_nl ! Read parameters controlling trajectories
  public :: finalize_traj_samp
  public :: init_traj_samp

  ! Defaults mean save all trajectories to file 'traj.bin'
  logical :: samp_flag   = .false.      ! Sample or not
  real(r_kind) :: prob   = 1           ! What proportion
  real(r_kind) :: min_vx = -1000 ! minimum x velocity
  real(r_kind) :: max_vx =  1000 ! maximum x velocity 
  real(r_kind) :: min_vy = -1000 ! minimum y velocity
  real(r_kind) :: max_vy =  1000 ! maximum y velocity 
  real(r_kind) :: min_t  = -1000 ! start time
  real(r_kind) :: max_t  =  1000 ! end time

  namelist /traj_samp_nl/ samp_flag, prob, min_vx, max_vx, min_vy, max_vy, &
       min_t, max_t

  integer traj_samp_fp

  integer :: total_written = 0 ! Count how many we write
  integer :: num_particle = 0  ! How many have we created


contains
  subroutine init_traj_samp(dirn)
    character(len=*), intent(in) :: dirn
    if (samp_flag) then
       write(std_out,*)'Writing trajectory data to ' &
            // trim(dirn) // 'traj.bin'
       call assign_lun(traj_samp_fp)
       call err_open(traj_samp_fp,file=trim(dirn) // 'traj.bin', &
            status='replace', form='unformatted')
    end if
  end subroutine init_traj_samp
  ! Read in the control parameters and open the output file
  subroutine read_traj_samp_nl(fp)
    integer, intent(in) :: fp
    integer iostat
    read(fp,nml=traj_samp_nl,err=1,iostat=iostat)
    return
1   call print_iostat(iostat,std_err,&
         'Error with traj_samp_nl reading file "'//trim(get_name(fp))//'"')
    call error_halt
  end subroutine read_traj_samp_nl
  subroutine write_traj_samp_nl(fp)
    integer, intent(in) :: fp
    write(fp,'(A)')       ' &traj_samp_nl'
    write(fp,'(A,L1)')    '    samp_flag = ', samp_flag
    write(fp,'(A,G11.4)')'    prob      = ', prob
    write(fp,'(A,G11.4)')'    min_vx    = ', min_vx
    write(fp,'(A,G11.4)')'    max_vx    = ', max_vx
    write(fp,'(A,G11.4)')'    min_vy    = ', min_vy
    write(fp,'(A,G11.4)')'    max_vy    = ', max_vy
    write(fp,'(A,G11.4)')'    min_t     = ', min_t
    write(fp,'(A,G11.4)')'    max_t     = ', max_t
    write(fp,'(A)')       ' /'
    write(fp,'(A)')' '
  end subroutine write_traj_samp_nl
  subroutine finalize_traj_samp()
    if (samp_flag) then
       call err_close(traj_samp_fp)
       
    end if
  end subroutine finalize_traj_samp
  function new_grain_num() result(n)
    integer n
    num_particle = num_particle+1
    n = num_particle
  end function new_grain_num
  subroutine add_grain(g,mom)
    use grain_var
    type(particles), intent(inout) :: g(:)
    logical, optional, intent(in) :: mom ! Momentum not velocity specified
    integer i,n
    n = size(g)
    ! Give each particle it's own number
    g%num = num_particle + (/(i,i=1,n)/)
    num_particle = num_particle+n 

    g%alive = .true.
    g%mass = grain_dens*g%diam**3*pi/6  ! Calculate mass
    g%ff = (pi * g%diam * air_viscosity) / (8*g%mass)
    g%rc = (air_density * g%diam) / air_viscosity
    if (present(mom)) then 
       if (mom) g%vel = g%vel/g%mass
    end if
    g%start_vel = g%vel
    g%start_pos = g%pos
    g%start_t = g%time
    g%max_y = g%pos%y    
    g%acc = null_rvec
  end subroutine add_grain
  subroutine delete_grain(g)
   type(particles), intent(in) :: g ! Just delete grains one at time
   real(r_kind) rand
!   call assert (g%alive,'traj_samp.f90 Deleting a dead grain')
   if (.not. samp_flag) return
   call random_number(rand)
   if (rand > prob) return
   if (g%start_vel%x > max_vx .or. g%start_vel%x < min_vx) return
   if (g%start_vel%y > max_vy .or. g%start_vel%y < min_vy) return
   if (g%start_t > max_t .or. g%start_t < min_t) return
   total_written = total_written + 1  
   write(traj_samp_fp)g%num, g%parent, g%origin, 0, g%start_t, g%time, &
        g%start_pos, g%pos, g%start_vel, g%vel, g%max_y
 end subroutine delete_grain
end module traj_samp
