! Functions for dealing with exact solution of boundary layer equation 
! d/dx[(nu + k^2x^2|du/dx|)du/dx]=0
module wall_law
  use precision
  use sysdep_io
  use maths
  implicit none
  private
  public :: wall_law_u
  public :: wall_law_du
  public :: calc_friction_vel
  public :: calc_wall_boundary

  ! Convergence tolerance
  real(r_kind), parameter :: con_tol2 = 1e-12
  real(r_kind), parameter :: con_tol  = 1e-6

  interface wall_law_u
     module procedure  wall_law_us, wall_law_ua
  end interface 
  interface wall_law_du
     module procedure  wall_law_dus, wall_law_dua
  end interface 
contains
  ! Given roughness height and friction velocity calculate 
  ! wind velocity at certain points
  
  ! Returns analytic velocity profile
  ! x is height from surface
  ! x0 is friction height
  ! fv is friction velocity
  ! k is von karmans constant
  ! myu is kinematic viscosity

  function wall_law_us(x,x0,fv,k,myu,no_slip) result(u)
    real(r_kind), intent(in) :: x   ! Height
    real(r_kind), intent(in) :: x0  ! Roughness height
    real(r_kind), intent(in) :: fv  ! Friction velocity
    real(r_kind), intent(in) :: k   ! Von karman's constant
    real(r_kind), intent(in) :: myu ! Kinematic viscosity
    logical,      intent(in) :: no_slip
    real(r_kind) :: u,y,y0
    if (abs(fv)<tiny(fv)) then
       u = 0
       return
    end if
    y = (2*k*abs(fv)/myu)*(x+x0)
    y0 = (2*k*abs(fv)/myu)*x0
    if (no_slip) then
       u = fv/k*(-y/(1+sqrt(1+y**2))+log(y+sqrt(1+y**2))&
            +y0/(1+sqrt(1+y0**2))-log(y0+sqrt(1+y0**2)))
    else
       u = fv/k*(2/(1+y+sqrt(1+y**2))+log((y+sqrt(1+y**2))/(2*y0)))
    end if
  end function wall_law_us

  function wall_law_ua(x,x0,fv,k,myu,no_slip) result(u)
    real(r_kind), intent(in) :: x(:)! Height
    real(r_kind), intent(in) :: x0  ! Roughness height
    real(r_kind), intent(in) :: fv  ! Friction velocity
    real(r_kind), intent(in) :: k   ! Von karman's constant
    real(r_kind), intent(in) :: myu ! Kinematic viscosity
    real(r_kind) :: u(size(x)),y(size(x)),y0
    logical     , intent(in) :: no_slip ! Boundary condition
    if (abs(fv)<tiny(fv)) then
       u = 0
       return
    end if
    y = (2*k*abs(fv)/myu)*(x+x0)
    y0 = (2*k*abs(fv)/myu)*x0
    if (no_slip) then
       u = fv/k*(-y/(1+sqrt(1+y**2))+log(y+sqrt(1+y**2))&
            +y0/(1+sqrt(1+y0**2))-log(y0+sqrt(1+y0**2)))
    else
       y0 = (2*k*abs(fv)/myu)*x0
       u = fv/k*(2/(1+y+sqrt(1+y**2))+log((y+sqrt(1+y**2))/(2*y0)))
    end if
  end function wall_law_ua
  function wall_law_dus(x,x0,fv,k,myu) result(u)
    real(r_kind), intent(in) :: x   ! Height
    real(r_kind), intent(in) :: x0  ! Roughness height
    real(r_kind), intent(in) :: fv  ! Friction velocity
    real(r_kind), intent(in) :: k   ! Von karman's constant
    real(r_kind), intent(in) :: myu ! Kinematic viscosity
    real(r_kind) :: u,y
    y = (2*k*fv/myu)*(x+x0)
    u = 2*fv**2/myu/(1+sqrt(1+y**2))
  end function wall_law_dus
  function wall_law_dua(x,x0,fv,k,myu) result(u)
    real(r_kind), intent(in) :: x(:)   ! Height
    real(r_kind), intent(in) :: x0  ! Roughness height
    real(r_kind), intent(in) :: fv  ! Friction velocity
    real(r_kind), intent(in) :: k   ! Von karman's constant
    real(r_kind), intent(in) :: myu ! Kinematic viscosity
    real(r_kind) :: u(size(x)),y(size(x))
    y = (2*k*fv/myu)*(x+x0)
    u = 2*fv**2/myu/(1+sqrt(1+y**2))
  end function wall_law_dua
  
  ! Calculate what friction velocity produces a velocity u at height y
  ! Uses interval bisection so takes around 55 iterations in double precision
  function calc_friction_vel(u,z,z0,k,myu,no_slip) result(fv)
    real(r_kind), intent(in) :: u   ! wind velocity
    real(r_kind), intent(in) :: z   ! at height z
    real(r_kind), intent(in) :: z0  ! friction height
    real(r_kind), intent(in) :: k   ! Von karman's constant
    real(r_kind), intent(in) :: myu ! Kinematic viscosity
    logical, intent(in)      :: no_slip ! Is boundary no slip condition
    real(r_kind) fv,x,y,a,sy,oy,nu,nz
    real(r_kind) z1,z2,u1,u2,c

    integer, parameter :: maxit = 20 ! Should take 3 or 4 iterations 
    integer n
    if (u<0) then
       fv=0
       return
    end if
    call assert(u>=0,'wall_law.f90: called with negative velocity')
    nz = z+z0

    if (no_slip) then
       z1 = z0
       z2 = z0+z
       u1 = 0
       u2 = u
       c = 2*k**2/myu
       x = c*u2-u1-1/z2+1/z1
       y = sqrt(c*(u2-u1)/(z2-z1))
       do n=1,maxit
          oy = y
          y = (x + sqrt(1+(z2*y)**2)/z2-sqrt(1+(z1*y)**2)/z1)&
               /(asinh(z2*y)-asinh(z1*y))
          if(abs(y-oy)<=con_tol) exit
       end do
       fv = myu*y/(2*k)
    else
       x = 2*k**2*nz*u/myu
       a = log(nz/(2*z0))
       y = x/(1+a+log(real(2,r_kind))) ! Initial guess

       ! Newton Raphson Iteration
       do n=1,maxit
          oy = y
          sy = sqrt(1+y**2)
          y = (x-2*y/(1+y+sy))/(log(1+sy/y)+a)
          if(abs(y-oy)<=con_tol) exit
       end do

       fv = myu*y/(2*k*nz)

       ! Check accuracy of solution
       sy = sqrt(1+y**2)
       nu = fv/k*(2/(1+y+sy)+log(1+sy/y)+a)

       if (abs(nu-u)>=con_tol .or. n>=maxit ) then
          write(std_err,'(1X,A,G12.4,A,I2,A)')&
               'Calculate friction velocity ',fv,' in ',&
               n,' iterations'
          write(std_err,*)'error',nu-u
          call error_halt(&
               'wall_law.f90: Friction velocity is inaccurate')
       end if
    end if

  end function calc_friction_vel

  ! Calculate boundary condition
  ! Given a velocity u at height z1
  ! What is the velocity at height z such that
  ! u(z) + (z1-z)*du(z)=u(z1)
  ! and u(z) and du(z) satisfy law of the wall with friction height z0
  subroutine calc_wall_boundary(nu,du,z,u,z1,z0,k,myu)
    real(r_kind), intent(out) :: nu  ! u(z)
    real(r_kind), intent(out) :: du  ! du(z)
    real(r_kind), intent(in)  :: z   ! calclulate at height z
    real(r_kind), intent(in)  :: u   ! u(z1)
    real(r_kind), intent(in)  :: z1  ! 
    real(r_kind), intent(in)  :: z0  ! friction height
    real(r_kind), intent(in)  :: k   ! Von karman's constant
    real(r_kind), intent(in)  :: myu ! Kinematic viscosity
    real(r_kind) y,a,b,c,e,sy,oy,nz,err(2),fv,ou
    integer, parameter :: maxit = 20 ! Should take 3 or 4 iterations 
    integer n
    logical neg

    nz = z+z0
    ou = abs(u)
    e = (z1-z)/nz

    c = 2*k*nz/myu

    a = log(nz/(2*z0))

    if (u==0) then
       nu=0
       du=0
       return
    else if (u<0) then
       neg=.true.
    else
       neg=.false.
    end if
 
    b = k*c*ou
 
    ! Make and refine intial guess
    y = b/(1+a+log(real(2,r_kind)))
    sy = sqrt(1+y**2)

    do n=1,maxit
       if (y<=0) call error_halt('wall.law.f90: Iteration has become negative')
       oy = y
       y = (b-2*y/(1+y+sy)+e*y**2/(sy*(1+sy)))/(log(1+sy/y)+a+e*y/sy)
       sy = sqrt(y**2+1)
       nu = y*(2/(1+y+sy) + log(1+sy/y)+a)/(c*k)
       du = y**2/(k*c*nz*(1+sy))
       if (abs(y-oy)<=con_tol2 &
            .and. abs(nu+du*e*nz-ou) <=con_tol) &
            exit
    end do

    if (neg) then
       nu = -nu
       du = -du
    end if

    if (n>=maxit ) call disp_err('maximum iterations exceeded, ')

    err = ar_error(nu+nz*e*du,u)
    if (err(2)>con_tol) call disp_err('nu+nz*e*du /= u, ')
    
    fv = ssqrt(du*(myu+k**2*nz**2*abs(du)))
    err=ar_error(nu,wall_law_u(z,z0,fv,k,myu,.false.),con_tol)
    if (err(2)>con_tol) call disp_err('nu /= f(du), ')

    contains
      subroutine disp_err(str)
        character(len=*), intent(in) :: str
       call error_msg(str//&
            '"wall_law.f90" function "calc_wall_boundary"')
       write(std_err,'(1X,3(A,G12.4),10x)')'z0=',z0,', k =',k, ', myu=',myu
       write(std_err,'(1X,3(A,G12.4),10x)')'z =',z ,', u =',u, ', z1 =',z1
       write(std_err,'(1X,3(A,G12.4),10x)')'nu=',nu,', du=',du,', y  =',y
       write(std_err,'(1X,A,2G12.4,10x)')'err=',err
       write(std_err,'(1X,A,I6,10x)')'Iterations',n
       call error_halt
     end subroutine disp_err
  end subroutine calc_wall_boundary
end module wall_law

