module wind
  use precision
  use sysdep_io
  use maths_const
  use discrete  
  use maths
  use locate
  use track_profile
  use stats
  use wall_law
  use ansi_strings
  use maths_fnct

  implicit none
  private 

  public :: read_wind_nl
  public :: write_wind_nl
  public :: read_wind_state    ! Reads wind velocity
  public :: write_wind_state   ! Write wind velocity
  public :: write_wind_profile ! Writes wind velocity and wind stress 
  public :: calc_wind          ! Time evolves wind velocity
  public :: init_wind          ! Initialize wind state
  public :: end_wind           ! Save wind state
  public :: air_vel            ! Return air velocity in space and time
  public :: get_b_stress       ! Get the wind stress at the ground
  public :: display_wind_step  ! Display info every step
  public :: write_wind_step    ! Write out to data file

  public :: air_viscosity
  public :: air_density

  ! Boundary conditions
  type(fnct_t)     :: boundary_f = null_fnct_t 
  type(fnct_pub_t) :: bound_fnct = fnct_pub_t("constant",0.4)
  logical          :: top_is_fv  = .true.  ! friction vel or velocity
  logical          :: bot_is_fv  = .false. ! friction vel or velocity
  logical          :: no_slip    = .false. ! no slip u=0
  logical          :: bot_exact  = .false. ! how to calculate bottom condition
  real(r_kind)     :: pressure   = 0       ! Pressure gradient

  namelist /wind_boundary_nl/ bound_fnct, top_is_fv, bot_is_fv, &
       no_slip, bot_exact, pressure

  integer      :: maxit         = 100         ! Maximum iterations
  integer      :: minit         = 1           ! Minimum iterations
  real(r_kind) :: eps           = 1e-6        ! Relative error tolerance
  real(r_kind) :: yzero         = 1e-5        ! Roughness height
  real(r_kind) :: air_density   = 1.225       ! Density
  real(r_kind) :: air_viscosity = 1.7763e-05  ! Viscosity
  real(r_kind) :: vkarman       = 0.41        ! Von Karman's constant
  !lambda=1-fully implicit, lambda=0-fully explicit, 0.5-Crank-Nicolson
  real(r_kind) :: lambda        = 1           ! Unimplimented

  logical      :: test_wind    = .false. ! Test accuracy of flow solver
  logical      :: dynamic_wind = .true.  ! solve time dependent equations
  logical      :: flat_wind    = .false. ! constant wind with height
  logical      :: const_wind   = .false. ! Is constant in time
  logical      :: load_state   = .false. ! Load state from a file
  character(len=FILE_NAME_LEN) :: wind_fn = "wind_state.bin" !state file name
  logical      :: three_point  = .false. ! Use three point stencil at the ends

  namelist /wind_nl/ maxit, minit, eps, lambda, three_point, &
       yzero, air_density, air_viscosity, vkarman, &
       wind_fn, test_wind, dynamic_wind, flat_wind, const_wind, load_state

  real(r_kind), allocatable :: stress(:)  ! Air stress 
  real(r_kind), allocatable :: windu(:)   ! Air velocity 
  real(r_kind), allocatable :: mix_len(:) ! Mixing length k^2y^2
  
  ! Coefficents for discetization formula at top and bottom
  real(r_kind) bot_corr ! Bottom correction
  real(r_kind) top_corr ! Top correction 

  real(r_kind) :: kvisco  ! Kinematic viscosity

  ! Now some other variables for reporting on the behaviour of the module
  type(stats_i_t) :: iter_stats ! How many iterations does convergence take
  type(stats_r_t) :: err_stats  ! Estimate size of max absolute error

  real(r_kind) :: max_u = 100 ! Maximum permitted wind speed
contains
  subroutine read_wind_nl(fp)
    integer, intent(in) :: fp
    integer iostat
    read(fp, nml=wind_boundary_nl,err=1,iostat=iostat)
    read(fp, nml=wind_nl,err=2,iostat=iostat)
    return
1   call print_iostat(iostat,std_err,&
         'Error with wind_boundary_nl reading file "'//trim(get_name(fp))//'"')
    call error_halt
    return
2   call print_iostat(iostat,std_err,&
         'Error with wind_nl reading file "'//trim(get_name(fp))//'"')
    call error_halt
  end subroutine read_wind_nl
  subroutine write_wind_nl(fp)
    integer, intent(in) :: fp
    write(fp,'(A)')      ' &wind_boundary_nl'
    write(fp,'(A)',advance='no')'    bound_fnct     = '
    call write_fnct(fp,bound_fnct)
    write(fp,'(A,L1)')   '    top_is_fv        = ', top_is_fv
    write(fp,'(A,L1)')   '    bot_is_fv        = ', bot_is_fv
    write(fp,'(A,L1)')   '    no_slip          = ', no_slip 
    write(fp,'(A,L1)')   '    bot_exact        = ', bot_exact 
    write(fp,'(A,G11.4)')'    pressure         = ', pressure
    write(fp,'(A)')       ' /'
    write(fp,'(A)')' '
    write(fp,'(A)')      ' &wind_nl'
    write(fp,'(A,I5)')   '    maxit            = ', maxit
    write(fp,'(A,I5)')   '    minit            = ', minit 
    write(fp,'(A,G11.4)')'    eps              = ', eps
    write(fp,'(A,G11.4)')'    yzero            = ', yzero 
    write(fp,'(A,G11.4)')'    air_density      = ', air_density
    write(fp,'(A,G11.4)')'    air_viscosity    = ', air_viscosity
    write(fp,'(A,G11.4)')'    vkarman          = ', vkarman
    write(fp,'(A,G11.4)')'    lambda           = ', lambda
    write(fp,'(A,L1)')   '    test_wind        = ', test_wind
    write(fp,'(A,L1)')   '    dynamic_wind     = ', dynamic_wind
    write(fp,'(A,L1)')   '    flat_wind        = ', flat_wind
    write(fp,'(A,L1)')   '    const_wind       = ', const_wind
    write(fp,'(A,L1)')   '    load_state       = ', load_state
    write(fp,'(A,L1)')   '    three_point      = ', three_point
    write(fp,'(A)')      '    wind_fn          = "' // trim(wind_fn) // '"'
    write(fp,'(A)')      ' /'
    write(fp,'(A)')' '
  end subroutine write_wind_nl
  ! Initialize the wind at time t  
  ! Iterate with no particle stress until wind speed converges 
  ! to steady state, for the boundary speed at time 
  subroutine end_wind(dirn)
    character(len=FILE_NAME_LEN), intent(in) :: dirn
    call write_wind_state(dirn)
    deallocate(stress, windu, mix_len)
    call stats_print(std_log, iter_stats,'Wind iterations')
    call stats_print(std_log, err_stats,'Wind errors')
    write(std_log,*)'Parameters'
    write(std_log,'(A,I5)')   '    maxit = ', maxit
    write(std_log,'(A,I5)')   '    minit = ', minit 
    write(std_log,'(A,G11.4)')'    eps   = ', eps
    load_state=.true.
  end subroutine end_wind
  subroutine init_wind(dirn)
    character(len=FILE_NAME_LEN), intent(in) :: dirn
    real(r_kind), parameter :: time = 0

    boundary_f = init_fnct(bound_fnct)
    call stats_clear(iter_stats)
    call stats_clear(err_stats)

    kvisco = air_viscosity/air_density

    allocate(stress(ngp), windu(0:ngr+1), mix_len(ngp))
    mix_len = vkarman**2*(grid_y+yzero)**2
    stress = 0
    windu  = 0

    ! Air velocity is defined on cell mid points
    ! These factors are for calculating second order accurate
    ! du/dy at top and bottom of grid
    ! Produces unstable algorithm on top grid point
    if (three_point) then
       bot_corr = 1 + (grid_my(1) -grid_y(1))/(grid_my(2) -grid_y(1))
       top_corr = 1 + (grid_y(ngp)-grid_my(ngr))/(grid_y(ngp)-grid_my(ngr-1))
    else
       top_corr = 1 
       bot_corr = 1 
    end if

    ! This is a very special case used for testing trajectory calculation
    ! The wind is taken to be constant everywhere
    ! The bottom stress is fixed
    if (flat_wind) then
       windu     = eval_fnct(boundary_f,time)
       stress    = 0
       stress(1) = air_density*windu(1)**2 ! Fix the bottom stress
       return
    end if

    ! If we're load_state load the old wind state
    if (load_state) then
       call read_wind_state(dirn) ! Load the old wind velocity
    else
       call calc_init_wind        ! calculate a wind state
    end if
    stress = calc_stress(windu) ! Calculate the stress
  end subroutine init_wind

  subroutine calc_init_wind()
    real(r_kind) :: time = 0

    real(r_kind) exactu(0:ngr+1)   ! What the exact wind speed should be
    real(r_kind) oldu(0:ngr+1) ! Wind speed before iteration

    real(r_kind) top_u  ! Top velocity
    real(r_kind) top_fv ! Friction velocity
    real(r_kind) err(2) ! For absolute and relative errors
    integer n

    ! Calculations assume that lower boundary condition enforces asymptotic
    ! equivalence to law of the wall.
    ! Analytic solution is wrong for no_slip boundary condition
    ! Otherwise we have to calculate the wind state
    ! First calculate or get  friction velocity
    if (top_is_fv) then
       top_fv = eval_fnct(boundary_f,time)
       top_u  = wall_law_u(grid_y(ngp),yzero,top_fv,vkarman,kvisco,no_slip)
    else  ! Solve for friction velocity that gives specified top velocity
       top_u  = eval_fnct(boundary_f,time)
       top_fv = calc_friction_vel(top_u,grid_y(ngp),yzero,vkarman,kvisco,no_slip)
    end if
    if (test_wind) then
       write(std_out,'(1X,3(A,F7.3))')&
            'Wind boundary: y = ',grid_y(ngp)+yzero,&
            ', u* = ', top_fv,&
            ', u(y) = ', top_u
    end if

    ! Now calculate exact velocity profile
    exactu = wall_law_u((/grid_y(1),grid_my,grid_y(ngp)/),yzero,top_fv,vkarman,kvisco,no_slip)
    call assert(exactu(ngp)-top_u,'Friction velocity calc error')
    windu = exactu

    ! Iterate the flow solver just to check it's stable
    do n = 1,5
       oldu = windu
       call calc_wind(time,huge(time))
    end do

    if (test_wind) then
       err = ar_error(windu,oldu) ! How much did it change on the last iteration?
       write(*,'(2(a,es11.4))')&
            ' Change,      absolute ',err(1),', relative ',err(2)
       err = ar_error(windu,exactu) ! How big are the errors?
       write(std_out,'(2(a,es11.4))')&
            ' Wind errors, absolute ',err(1),', relative ',err(2)
       if (err(2)>0.1) then
          write(std_out,'(1X,a,4(G11.4))')'bot windu  ',windu(0:3)
          write(std_out,'(1X,a,4(G11.4))')'bot exactu ',exactu(0:3)
          write(std_out,'(1X,a,4(G11.4))')'top windu  ',windu(ngp-3:ngp)
          write(std_out,'(1X,a,4(G11.4))')'top exactu ',exactu(ngp-3:ngp)
       end if
    end if
  end subroutine calc_init_wind
  subroutine read_wind_state(dirn)
    character(len=*), intent(in) :: dirn
    integer fp
    integer iostat
    call assign_lun(fp)
    call err_open(fp,file=trim(dirn)//trim(wind_fn),form='unformatted',&
         status='old',action='read')
    read(fp,err=1,iostat=iostat)windu(1:ngr)
    return
1   call print_iostat(iostat,std_err,'Error reading file "'//trim(get_name(fp))//'"')
    call error_halt
    call err_close(fp);
  end subroutine read_wind_state
  subroutine write_wind_state(dirn)
    character(len=*), intent(in) :: dirn
    integer fp
    call assign_lun(fp)
    call err_open(fp,file=trim(dirn)//trim(wind_fn),form='unformatted',&
         status='replace', action='write')
    write(fp)windu(1:ngr)
    call err_close(fp);
  end subroutine write_wind_state
  subroutine write_wind_profile(fp)
    integer, intent(in) :: fp
    write(fp)windu(1:ngr), stress
  end subroutine write_wind_profile
  ! Calculate the evolution of the wind from time to time+t_step
  subroutine calc_wind(time,t_step)
    use bed
    real(r_kind), intent(in) :: time
    real(r_kind), intent(in) :: t_step

    real(r_kind) norm       ! Measure of solution error
    real(r_kind) err(2)
    integer n               ! Number of iteration

    real(r_kind) ap(ngr)    ! Diagonal of tridiagonal system
    real(r_kind) rhs(ngr)   ! Right hand side of tridiagonal system
    real(r_kind) brhs(ngr)  ! Base rhs of tridiagonal system
    real(r_kind) du(ngp)    ! du/dy on grid points
    real(r_kind) eviss(ngp) ! turbulent viscosity
    real(r_kind) residual(0:ngr) ! Solution error

    real(r_kind) top_u      ! Top boundary velocity
    real(r_kind) top_du     ! Top boundary shear
    real(r_kind) top_fv     ! Top boundary friction velocity
    real(r_kind) bot_u      ! Bottom boundary velocity
    real(r_kind) bot_du     ! Bottom boundary shear
    real(r_kind) bot_fv     ! Bottom boundary friction velocity
    real(r_kind) oldu(0:ngr+1) ! Wind velocity at start of time step
    real(r_kind) bot_u1,bot_u2
    real(r_kind) bot_dfv,bot_y

    oldu = windu ! Original wind velocity

    if (flat_wind) then ! Wind is constant with height
       if (top_is_fv) then
          top_fv = eval_fnct(boundary_f,time)
          top_u  = wall_law_u(grid_y(ngp),yzero,top_fv,vkarman,kvisco,no_slip)
       else
          top_u  = eval_fnct(boundary_f,time)
          top_fv = calc_friction_vel(top_u,grid_y(ngp),yzero,vkarman,kvisco,no_slip)
       end if
       windu     = top_u                 ! Constant wind everywhere
       stress    = 0                     ! No stress
       stress(1) = air_density*top_fv**2 ! Fix bottom stress
       return  ! Finish 
    end if

    if (const_wind) then ! Wind is constant in time
       return
    end if

    brhs = - get_part_force(.not.dynamic_wind)/(air_density*grid_area) &
         + pressure/air_density*grid_dy

    ! Set Right Hand Side for equations
    if (dynamic_wind) then
       brhs = brhs + windu(1:ngr)*grid_dy/t_step 
    end if

    ! Calculate normal velocity gradient du
    du  = calc_shear(windu)
    ! We iterate the flow solver and average the value of du used
    ! in calculating the viscosity.
    ! A minimum number of iterations is needed to converge to the solution

    if (top_is_fv) then
       top_fv = eval_fnct(boundary_f,time)
       top_du = fv_to_du(top_fv,ngp)
    else
       top_u  = eval_fnct(boundary_f,time)
    end if
    norm  = 1   ! Size of error
    n = 0
    do ! Iterate flow solver
       where(abs(windu)>max_u)windu=sign(max_u,windu)
       n = n+1
       ! The bottom condition can be no slip or a mixed condition
       if (no_slip) then
          bot_u = 0
       else 
          call calc_wall_boundary(bot_u,bot_du,grid_y(1),&
               windu(1),grid_my(1), yzero,vkarman,kvisco)
          bot_fv = du_to_fv(bot_du,1)
          bot_y  = 2*vkarman*(grid_y(2)+yzero)*bot_fv/kvisco
          if (bot_y>0) then
             bot_dfv  = kvisco*bot_y/(grid_y(2)+yzero)&
                  /(asinh(bot_y)-log(2*bot_y*yzero/(grid_y(2)+yzero)))
          else ! Impose a no slip condition
             bot_u = 0
          end if
       end if
       windu(0)   = bot_u
       
       if (.not. top_is_fv) windu(ngp) = top_u 
       
       ! Has solution converged?
       if (n>=maxit .or. (n >= minit .and. norm < eps)) exit 
       
       ! Average with previous iteration for stability
       du = (du+calc_shear(windu))/2 ! Calculate shear velocity 
       if (top_is_fv) du(ngp) = top_du ! Impose boundary conditions

       eviss=(kvisco+mix_len*abs(du))/grid_mdy! Calculate Effective viscosity

       rhs = brhs ! Right hand side of solver

       if (top_is_fv) then
          rhs(ngr)   = rhs(ngr) + top_fv**2
          eviss(ngp) = 0
       else
          rhs(ngr)   = rhs(ngr) + eviss(ngp)*top_u
       end if

       if (no_slip .or. bot_y<=0) then
          rhs(1)     = rhs(1)   + eviss(1)*bot_u
       else ! Mixed condition solved by Newton-Raphson
          rhs(1)     = rhs(1)   - (bot_fv**2-bot_dfv*windu(1))
          eviss(1)   = bot_dfv
       end if
       
       ap  = eviss(2:ngp) + eviss(1:ngp-1)
       
       ! Add in time dependent terms
       if (dynamic_wind) then
          ap  = ap + grid_dy/t_step
       end if

       ! Calculate the residual solution error
       residual(1:ngr) = &
            tridiagmult(windu(1:ngr),-eviss(2:ngr),ap,-eviss(2:ngr))-rhs
       ! Boundary condition error
       if (no_slip .or. windu(1)<=0) then
          residual(0) = windu(0)
       else
          residual(0) = windu(0)&
               -wall_law_u(grid_y(1),yzero,du_to_fv(du(1),1),vkarman,kvisco,no_slip)
       end if
       err = (/maxval(abs(residual)), sqrt(sum(residual**2)/ngr)/) 
       norm=err(1) ! Iteration criteria is max absolute error
       if (test_wind .and. n>maxit/2) then
          write(std_err,*)' Iteration',n,' -----------------------------------'
          write(std_err,'(1X,a,4(1X,G11.4),10x)')'bot windu',windu(0:2)
          write(std_err,'(1X,a,4(1X,G11.4),10x)')'top windu',windu(ngp-2:ngp)
          write(std_err,'(1X,a,3(1X,G11.4),10x)')'bot residual',residual(0:2)
          write(std_err,'(1X,a,3(1X,G11.4),10x)')'top residual',residual(ngr-2:ngr)
       end if

       ! Calculate a new solution
       ! Bottom error should compare du and u at x=x0

       call tridiagsolv(windu(1:ngr),-eviss(2:ngr),ap,-eviss(2:ngr),rhs)
    end do ! iterate solution loop
    if (top_is_fv) then
       windu(ngp) = windu(ngr) + grid_mdy(ngp)/top_corr*(top_du &
            - (1-top_corr)*du(ngp-1))
    end if
    if (n >= maxit) then ! Failed to converge
       write(std_err,"(1X,A,2G12.4,A,10X)")STR_RED//&
            "wind: Max iterations exceeded."&
            //" error ",err,STR_RESET
       if (test_wind) then
          write(std_out,'(1X,a,4(1X,G11.4))')'bot windu',windu(0:2)
          write(std_out,'(1X,a,4(1X,G11.4))')'top windu',windu(ngp-2:ngp)
          write(std_out,'(1X,a,3(1X,G11.4))')'bot residual',residual(0:2)
          write(std_out,'(1X,a,3(1X,G11.4))')'top residual',residual(ngr-2:ngr)
       end if
    end if
    ! Count statistics on the functioning of the algorithm
    call stats_add(iter_stats,n)
    call stats_add(err_stats,norm)
    if(.not.dynamic_wind) windu = (windu+oldu)/2 ! Average wind

    stress = calc_stress(windu) ! Now calculate the stress on the grid points
  end subroutine calc_wind
  function get_b_stress() result(r)
    real(r_kind) r
    r = stress(1)
  end function get_b_stress
  subroutine display_wind_step(fp)
    integer, intent(in) :: fp
    real(r_kind) az0,bz0,ut,ub,du
    ub = ssqrt(stress(1)/air_density)
    ut = ssqrt(stress(ngp)/air_density)
    if (ut>0) then
       az0 = (grid_y(ngp) + yzero)* exp(-vkarman*windu(ngp)/ut)
    else
       az0 = 0
    end if
    if (windu(ngr) == windu(ngr-1)) then
       bz0=0
    else
       du = (windu(ngr) + windu(ngr-1))/(2*(windu(ngr) - windu(ngr-1)))
       bz0 = sqrt((grid_my(ngr-1)+ yzero)*(grid_my(ngr)+ yzero)) &
            * ((grid_my(ngr-1)+ yzero)/(grid_my(ngr)+ yzero))**du
    end if

    write(fp,'(1X,2(A,F4.2),2(A,F6.2),A,2(ES9.2))')&
         "u*: bottom ", ub ,", top ", ut, &
         " u: bottom ", windu(0) ,", top ", windu(ngp), &
         ", z0 ", az0, bz0
  end subroutine display_wind_step
  subroutine write_wind_step(fp)
    integer, intent(in) :: fp
    real(r_kind) bz0,ut,ub,du
    ub = ssqrt(stress(1)/air_density)
    ut = ssqrt(abs(stress(ngp))/air_density)
    if (windu(ngr) == windu(ngr-1)) then
       bz0=0
    else
       du = (windu(ngr) + windu(ngr-1))/(2*(windu(ngr) - windu(ngr-1)))
       bz0 = sqrt((grid_my(ngr-1)+ yzero)*(grid_my(ngr)+ yzero)) &
            * ((grid_my(ngr-1)+ yzero)/(grid_my(ngr)+ yzero))**du
    end if

    write(fp,'(4(F8.3),1(ES9.2))',advance='no')ub,ut,windu(0),windu(ngp),bz0
  end subroutine write_wind_step
  ! Calculate the stress for a particular velocity distribution
  function calc_stress(u) result(s)
    real(r_kind), intent(in) :: u(0:ngp)
    real(r_kind) s(1:ngp)
    real(r_kind) du(1:ngp)
    du=calc_shear(u)
    s=air_density*du*(kvisco+mix_len*abs(du)) 
  end  function calc_stress   
  function calc_shear(u) result(du)
    real(r_kind), intent(in) :: u(0:ngp)
    real(r_kind) du(1:ngp)
    du      = (u(1:ngr+1)-u(0:ngr))/grid_mdy
    du(1)   = bot_corr*du(1)   + (1-bot_corr)*du(2)
    du(ngp) = top_corr*du(ngp) + (1-top_corr)*du(ngp-1)
  end function calc_shear
  ! return the air velocity at time t height y with possible position j
  function air_vel(y,t,j) result(v)
    real(r_kind), intent(in) :: y            ! Current height
    real(r_kind), intent(in) :: t            ! Time
    real(r_kind), intent(in), optional :: j  ! Grid position
    real(r_kind) :: v                        ! Velocity
    integer, save  :: i=1
    if (present(j)) then
       i=j
    else
       i = lookup(y,grid_my,i) ! Mostly we are searching in similar places
    end if
    if (i==0) then
       v = (windu(1)*(y-grid_y(1))+windu(0)*(grid_my(1)-y))/(grid_my(1)-grid_y(1))
    else if (i==ngr) then
       v = (windu(ngr+1)*(y-grid_my(ngr))+windu(ngr)*(grid_y(ngr+1)-y))&
            /(grid_y(ngr+1)-grid_my(ngr))
    else
       v = (windu(i)*(y-grid_my(i))+windu(i+1)*(grid_my(i+1)-y))/grid_mdy(i+1)
    end if
  end function air_vel

  ! Prevent negative velocities but preserve momentum
  subroutine remove_neg(u)
    real(r_kind), intent(inout) :: u(1:ngr)
    integer i
    do i = 1, ngr-1 
       if (u(i)<0) then
          u(i+1) = u(i+1) + u(i)*grid_dy(i)/grid_dy(i+1)
          u(i) = 0
       else 
          exit
       end if
    end do
    if (i>1) write(std_out,*)"negative velocities detected"
    if (u(ngr)<0) then
       write(std_log,*)"negative top velocity detected. dumping windu"
       write(std_log,*)u
       call error_halt("wind.f90: top velocity negative")
    end if
  end subroutine remove_neg
  ! Calculate friction velocity from shear velocity
  function fv_to_du(fv,n) result(du)
    real(r_kind), intent(in) :: fv ! Friction velocity
    integer,      intent(in) :: n  ! Grid point
    real(r_kind) du
    du = sign(2*fv**2/(kvisco+sqrt(kvisco**2+4*fv**2*mix_len(n))),fv)
  end function fv_to_du
  ! Calculate shear velocity from friction velocity
  function du_to_fv(du,n) result(fv)
    real(r_kind), intent(in) :: du ! Shear velocity
    integer,      intent(in) :: n  ! Grid point
    real(r_kind) fv
    fv = ssqrt(du*(kvisco+mix_len(n)*abs(du)))
  end function du_to_fv
end module wind
  
    ! The top boundary condition can be friction velocity or absolute velocity
!!$             bot_u2=bot_u1
!!$             bot_u1=bot_u
!!$             if (mod(n,3)==0) bot_u = aitken(bot_u2,bot_u1,bot_u)
